@@ -1,18 +1,14 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_CONTENT_H_
 #define AWKWARD_CONTENT_H_
 
-#include <cstdio>
 #include <map>
 
-#include "awkward/cpu-kernels/util.h"
+#include "awkward/common.h"
 #include "awkward/Identities.h"
 #include "awkward/Slice.h"
-#include "awkward/io/json.h"
-#include "awkward/type/Type.h"
 #include "awkward/Index.h"
-#include "awkward/Reducer.h"
 
 namespace awkward {
   class Content;
@@ -20,6 +16,17 @@ namespace awkward {
   using ContentPtrVec = std::vector<std::shared_ptr<Content>>;
   class Form;
   using FormPtr       = std::shared_ptr<Form>;
+  using FormKey       = std::shared_ptr<std::string>;
+  class ArrayCache;
+  using ArrayCachePtr = std::shared_ptr<ArrayCache>;
+  class Type;
+  using TypePtr = std::shared_ptr<Type>;
+  class Reducer;
+  class ToJson;
+  class ToJsonPrettyString;
+  class ToJsonString;
+  class ToJsonPrettyFile;
+  class ToJsonFile;
 
   /// @class Form
   ///
@@ -29,13 +36,20 @@ namespace awkward {
   ///
   /// Forms may be thought of as low-level types, whereas Type is a high-level
   /// type. There is a one-to-many relationship from Type to Form.
-  class EXPORT_SYMBOL Form {
+  class LIBAWKWARD_EXPORT_SYMBOL Form {
   public:
-    static FormPtr fromjson(const std::string& data);
+    static FormPtr
+      fromnumpy(char kind,
+                int64_t itemsize,
+                const std::vector<int64_t>& inner_shape);
+    static FormPtr
+      fromjson(const std::string& data);
 
-    /// @brief Called by subclass constructors; assigns #has_identities
-    /// and #parameters upon construction.
-    Form(bool has_identities, const util::Parameters& parameters);
+    /// @brief Called by subclass constructors; assigns #has_identities,
+    /// #parameters, and #form_key upon construction.
+    Form(bool has_identities,
+         const util::Parameters& parameters,
+         const FormKey& form_key);
 
     /// @brief Empty destructor; required for some C++ reason.
     virtual ~Form() { }
@@ -57,12 +71,33 @@ namespace awkward {
     virtual const FormPtr
       shallow_copy() const = 0;
 
+    /// @brief Copies this node, adding or replacing a form_key.
+    virtual const FormPtr
+      with_form_key(const FormKey& form_key) const = 0;
+
     /// @brief Returns `true` if this Form is equal to the other Form; `false`
     /// otherwise.
+    ///
+    /// @param check_identities If `true`, Forms are not equal unless they both
+    /// #has_identities.
+    /// @param check_parameters If `true`, Forms are not equal unless they have
+    /// the same #parameters.
+    /// @param check_form_key If `true`, Forms are not equal unless they have
+    /// the same #form_key.
+    /// @param compatibility_check If `true`, this is part of a compatibility
+    /// check between an expected Form (`this`) and a generated array's Form
+    /// (`other`). When the expected Form is a VirtualForm, it's allowed to be
+    /// less specific than the `other` VirtualForm.
     virtual bool
       equal(const FormPtr& other,
             bool check_identities,
-            bool check_parameters) const = 0;
+            bool check_parameters,
+            bool check_form_key,
+            bool compatibility_check) const = 0;
+
+    /// @brief Returns `true` if this Form has the same #form_key as the other.
+    bool
+      form_key_equals(const FormKey& other_form_key) const;
 
     /// @brief The parameter associated with `key` at the first level
     /// that has a non-null value, descending only as deep as the first
@@ -96,6 +131,16 @@ namespace awkward {
     virtual int64_t
       purelist_depth() const = 0;
 
+    /// @brief Returns `true` if this dimension has option-type; `false`
+    /// otherwise.
+    virtual bool
+      dimension_optiontype() const = 0;
+
+    /// @brief An optional string associated with this Form, usually specifying
+    /// where an array may be fetched.
+    const FormKey
+      form_key() const;
+
     /// @brief Returns (a) the minimum list-depth and (b) the maximum
     /// list-depth of the array, which can differ if this array "branches"
     /// (differs when followed through different fields of a RecordForm or
@@ -177,28 +222,54 @@ namespace awkward {
 
     /// @brief Internal function for adding identities in #tojson.
     ///
-    /// Must be called between `builder.beginrecord()` and `builder.endrecord`.
+    /// Must be called between `builder.beginrecord()` and `builder.endrecord()`.
     void
       identities_tojson(ToJson& builder, bool verbose) const;
 
     /// @brief Internal function for adding parameters in #tojson.
     ///
-    /// Must be called between `builder.beginrecord()` and `builder.endrecord`.
+    /// Must be called between `builder.beginrecord()` and `builder.endrecord()`.
     void
       parameters_tojson(ToJson& builder, bool verbose) const;
 
-    protected:
-    /// @brief See #has_identities;
+    /// @brief Internal function for adding form_key in #tojson.
+    ///
+    /// Must be called between `builder.beginrecord()` and `builder.endrecord()`.
+    void
+      form_key_tojson(ToJson& builder, bool verbose) const;
+
+    /// @brief Returns the Form that would result from a range-slice.
+    ///
+    /// Matches the operation of Content#getitem_range.
+    virtual const FormPtr
+      getitem_range() const;
+
+    /// @brief Returns the Form that would result from a field-slice.
+    ///
+    /// Matches the operation of Content#getitem_field.
+    virtual const FormPtr
+      getitem_field(const std::string& key) const = 0;
+
+    /// @brief Returns the Form that would result from a fields-slice.
+    ///
+    /// Matches the operation of Content#getitem_fields.
+    virtual const FormPtr
+      getitem_fields(const std::vector<std::string>& keys) const = 0;
+
+  protected:
+    /// @brief See #has_identities
     bool has_identities_;
-    /// @brief See #parameters;
+    /// @brief See #parameters
     util::Parameters parameters_;
+    /// @brief See #form_key
+    FormKey form_key_;
   };
 
   /// @class Content
   ///
   /// @brief Abstract superclass of all array node types (flat hierarchy).
   /// Any Content can be nested within any other Content.
-  class EXPORT_SYMBOL Content {
+  class LIBAWKWARD_EXPORT_SYMBOL Content {
   public:
     /// @brief Called by all subclass constructors; assigns #identities and
     /// #parameters upon construction.
@@ -281,14 +352,17 @@ namespace awkward {
     virtual const FormPtr
       form(bool materialize) const = 0;
 
-    /// @brief If true, the array contains a VirtualArray with an unknown Form.
-    virtual bool
-      has_virtual_form() const = 0;
+    /// @brief Returns the kernel library enum for all nested `ptr_lib`
+    /// within the array's tree structure. If different libraries are
+    /// associated with nodes of the same tree, this returns `kernel::lib::size`
+    /// to indicate that it is "mixed".
+    virtual kernel::lib
+      kernels() const = 0;
 
-    /// @brief If true, the array contains a VirtualArray with an unknown
-    /// length.
-    virtual bool
-      has_virtual_length() const = 0;
+    /// @brief Accumulates all the unique #ArrayCache objects from nested
+    /// #VirtualArray nodes. (Uniqueness is determined by pointer value.)
+    virtual void
+      caches(std::vector<ArrayCachePtr>& out) const = 0;
 
     /// @brief Internal function to build an output string for #tostring.
     ///
@@ -401,11 +475,19 @@ namespace awkward {
     virtual const ContentPtr
       getitem_field(const std::string& key) const = 0;
 
+    virtual const ContentPtr
+      getitem_field(const std::string& key,
+                    const Slice& only_fields) const = 0;
+
     /// @brief This array with the first nested RecordArray replaced by
     /// a RecordArray of a given subset of `keys`.
     virtual const ContentPtr
       getitem_fields(const std::vector<std::string>& keys) const = 0;
 
+    virtual const ContentPtr
+      getitem_fields(const std::vector<std::string>& keys,
+                     const Slice& only_fields) const = 0;
+
     /// @brief Entry point for general slicing: Slice represents a tuple of
     /// SliceItem nodes applying to each level of nested lists.
     virtual const ContentPtr
@@ -469,7 +551,7 @@ namespace awkward {
     /// {@link ListArrayOf ListArray}, only the `starts` and `stops` get
     /// duplicated, not the `content` (and similarly for all other node types).
     virtual const ContentPtr
-      carry(const Index64& carry) const = 0;
+      carry(const Index64& carry, bool allow_lazy) const = 0;
 
     /// @brief The parameter associated with `key` at the first level
     /// that has a non-null value, descending only as deep as the first
@@ -482,6 +564,11 @@ namespace awkward {
     bool
       purelist_isregular() const;
 
+    /// @brief Returns `true` if this dimension has option-type; `false`
+    /// otherwise.
+    virtual bool
+      dimension_optiontype() const;
+
     /// @brief The list-depth of this array, not counting any contained
     /// within a RecordArray.
     ///
@@ -491,14 +578,14 @@ namespace awkward {
     /// If this array contains a {@link UnionArrayOf UnionArray} with
     /// different depths, the return value is `-1`.
     virtual int64_t
-      purelist_depth() const;
+      purelist_depth() const = 0;
 
     /// @brief Returns (a) the minimum list-depth and (b) the maximum
     /// list-depth of the array, which can differ if this array "branches"
     /// (differs when followed through different fields of a RecordArray or
     /// {@link UnionArrayOf UnionArray}).
     virtual const std::pair<int64_t, int64_t>
-      minmax_depth() const;
+      minmax_depth() const = 0;
 
     /// @brief Returns (a) whether the list-depth of this array "branches,"
     /// or differs when followed through different fields of a RecordArray or
@@ -507,7 +594,7 @@ namespace awkward {
     /// If the array does not contain any records or heterogeneous data, the
     /// `first` element is always `true` and the `second` is simply the depth.
     virtual const std::pair<bool, int64_t>
-      branch_depth() const;
+      branch_depth() const = 0;
 
     /// @brief The number of fields in the first nested tuple or
     /// records or `-1` if this array does not contain a RecordArray.
@@ -593,10 +680,54 @@ namespace awkward {
     virtual bool
       mergeable(const ContentPtr& other, bool mergebool) const = 0;
 
+    /// @brief Returns `true` if this array has all the same buffers and
+    /// parameters as `other`; `false` otherwise.
+    ///
+    /// @param other The array to compare this with.
+    virtual bool
+      referentially_equal(const ContentPtr& other) const = 0;
+
+    /// @brief Partitions `this` array plus a list of `others` into a `head`
+    /// sequence and a `tail` sequence:
+    ///
+    /// `[this] + others == head + tail`
+    ///
+    /// The order is preserved and no arrays are lost. The `others` must
+    /// have at least one array, and `this` will be added to `head`,
+    /// though `tail` might be empty.
+    ///
+    /// (If `tail` is empty, `head` must contain at least two arrays!)
+    ///
+    /// The position of the split between `head` and `tail` is such that the
+    /// first array in `tail` must be merged with #reverse_merge (usually
+    /// because it is an option-type or union-type array, though this
+    /// rule depends on whether `this` is option-type, union-type, or neither).
+    ///
+    /// The purpose of this function is to prepare a `head` of arrays that
+    /// can all be merged "normally," followed by a `tail` whose first array
+    /// must be "reverse merged" and subsequent arrays recurse.
+    ///
+    /// This is the first step in #merge.
+    virtual const std::pair<ContentPtrVec, ContentPtrVec>
+      merging_strategy(const ContentPtrVec& others) const;
+
+    /// @brief Merges a single `other` with this array in reverse order:
+    /// `other` first, this last.
+    ///
+    /// Only arrays that need to be reversible have this function:
+    /// option-type and union-type arrays. Others raise a runtime error.
+    virtual const ContentPtr
+      reverse_merge(const ContentPtr& other) const;
+
     /// @brief An array with this and the `other` concatenated (this
     /// first, `other` last).
+    const ContentPtr
+      merge(const ContentPtr& other) const;
+
+    /// @brief Returns an array with this and the `others` concatenated
+    /// (in order, this first, `others` last).
     virtual const ContentPtr
-      merge(const ContentPtr& other) const = 0;
+      mergemany(const ContentPtrVec& others) const = 0;
 
     /// @brief Converts this array into a SliceItem that can be used in
     /// getitem.
@@ -654,7 +785,13 @@ namespace awkward {
     /// @param starts Staring positions of each group to combine as an
     /// {@link IndexOf Index}. These are downward pointers from an outer
     /// structure into this structure with the same meaning as in
-    /// {@link ListArrayOf ListArray}.
+    /// {@link ListArrayOf ListArray}. Only used by a reducer that
+    /// {@link Reducer#returns_positions returns_positions} (currently
+    /// only argmin and argmax).
+    /// @param shifts Per-element adjustments for any reducer that
+    /// {@link Reducer#returns_positions returns_positions}. Allows
+    /// for variable-length lists with axis != -1 and for missing values
+    /// (None) in argmin and argmax.
     /// @param parents Groups to combine as an {@link IndexOf Index} of
     /// upward pointers from this structure to the outer structure to reduce.
     /// @param outlength The length of the reduced array, after the operation
@@ -670,11 +807,49 @@ namespace awkward {
       reduce_next(const Reducer& reducer,
                   int64_t negaxis,
                   const Index64& starts,
+                  const Index64& shifts,
                   const Index64& parents,
                   int64_t outlength,
                   bool mask,
                   bool keepdims) const = 0;
 
+    /// @brief This array with one axis removed by applying a Reducer
+    ///
+    /// The user's entry point for this operation is #sort.
+    ///
+    /// @param negaxis The negative axis: `-axis`. That is, `negaxis = 1`
+    /// means the deepest axis level.
+    /// @param starts Staring positions of each group to combine as an
+    /// {@link IndexOf Index}. These are downward pointers from an outer
+    /// structure into this structure with the same meaning as in
+    /// {@link ListArrayOf ListArray}.
+    /// @param parents Groups to combine as an {@link IndexOf Index} of
+    /// upward pointers from this structure to the outer structure to reduce.
+    /// @param outlength The length of the array, after the operation
+    /// completes.
+    /// @param ascending If `true`, the values will be sorted in an ascending
+    /// order.
+    /// @param stable If `true`, the values will be sorted by a
+    /// stable sort algorithm to maintain the relative order of records with
+    /// equal keys (i.e. values).
+    virtual const ContentPtr
+      sort_next(int64_t negaxis,
+                const Index64& starts,
+                const Index64& parents,
+                int64_t outlength,
+                bool ascending,
+                bool stable,
+                bool keepdims) const = 0;
+
+    virtual const ContentPtr
+      argsort_next(int64_t negaxis,
+                   const Index64& starts,
+                   const Index64& parents,
+                   int64_t outlength,
+                   bool ascending,
+                   bool stable,
+                   bool keepdims) const = 0;
+
     /// @brief A (possibly nested) array of integers indicating the
     /// positions of elements within each nested list.
     ///
@@ -782,13 +957,24 @@ namespace awkward {
     /// If `false`, return a compact representation.
     /// @param maxdecimals Maximum number of decimals for floating-point
     /// numbers or `-1` for no limit.
+    /// @param nan_string user-defined string for a not-a-number (NaN) value
+    /// representation in JSON format
+    /// @param infinity_string user-defined string for a positive infinity
+    /// representation in JSON format
+    /// @param minus_infinity_string user-defined string for a negative
+    /// infinity representation in JSON format
     ///
     /// Although the JSON output contains all of the data from the array
     /// (and therefore might not be safe to print to the screen), it
     /// does not preserve the type information of an array. In particular,
     /// the distinction between ListType and RegularType is lost.
     const std::string
-      tojson(bool pretty, int64_t maxdecimals) const;
+      tojson(bool pretty, int64_t maxdecimals,
+             const char* nan_string = nullptr,
+             const char* infinity_string = nullptr,
+             const char* minus_infinity_string = nullptr,
+             const char* complex_real_string = nullptr,
+             const char* complex_imag_string = nullptr) const;
 
     /// @brief Writes a JSON representation of this array to a `destination`
     /// file.
@@ -799,6 +985,12 @@ namespace awkward {
     /// @param maxdecimals Maximum number of decimals for floating-point
     /// numbers or `-1` for no limit.
     /// @param buffersize Size of a temporary buffer in bytes.
+    /// @param nan_string user-defined string for a not-a-number (NaN) value
+    /// representation in JSON format
+    /// @param infinity_string user-defined string for a positive infinity
+    /// representation in JSON format
+    /// @param minus_infinity_string user-defined string for a negative
+    /// infinity representation in JSON format
     ///
     /// Although the JSON output contains all of the data from the array
     /// (and therefore might not be safe to print to the screen), it
@@ -808,7 +1000,12 @@ namespace awkward {
       tojson(FILE* destination,
              bool pretty,
              int64_t maxdecimals,
-             int64_t buffersize) const;
+             int64_t buffersize,
+             const char* nan_string = nullptr,
+             const char* infinity_string = nullptr,
+             const char* minus_infinity_string = nullptr,
+             const char* complex_real_string = nullptr,
+             const char* complex_imag_string = nullptr) const;
 
     /// @brief The number of bytes contained in all array buffers,
     /// {@link IndexOf Index} buffers, and Identities buffers, not including
@@ -838,6 +1035,37 @@ namespace awkward {
              bool mask,
              bool keepdims) const;
 
+    /// @brief This array with one axis sorted by applying a sorting algorithm
+    ///
+    /// This operation is implemented on each node through #sort_next.
+    ///
+    /// @param axis The axis to sort.
+    /// Negative `axis` counts backward from the deepest levels (`-1` is
+    /// the last valid `axis`).
+    /// @param ascending If `true`, the values will be sorted in an ascending
+    /// order.
+    /// @param stable If `true`, the values will be sorted by a
+    /// stable sort algorithm to maintain the relative order of records with
+    /// equal keys (i.e. values).
+    const ContentPtr
+      sort(int64_t axis, bool ascending, bool stable) const;
+
+    /// @brief This array indices with one axis sorted by applying
+    ///        a sorting algorithm
+    ///
+    /// This operation is implemented on each node through #argsort_next.
+    ///
+    /// @param axis The axis to sort.
+    /// Negative `axis` counts backward from the deepest levels (`-1` is
+    /// the last valid `axis`).
+    /// @param ascending If `true`, the values will be sorted in an ascending
+    /// order.
+    /// @param stable If `true`, the values will be sorted by a
+    /// stable sort algorithm to maintain the relative order of records with
+    /// equal keys (i.e. values).
+    const ContentPtr
+      argsort(int64_t axis, bool ascending, bool stable) const;
+
     /// @brief String-to-JSON map that augments the meaning of this
     /// array.
     ///
@@ -906,8 +1134,11 @@ namespace awkward {
     /// Equality is checked at the level of JSON DOMs. The `value` does not
     /// need to be exactly the same string; it needs to have equivalent JSON
     /// value.
+    ///
+    /// If `check_all`, every parameter is checked; otherwise, only
+    /// `"__array__"` and `"__record__"` are checked.
     bool
-      parameters_equal(const util::Parameters& other) const;
+      parameters_equal(const util::Parameters& other, bool check_all) const;
 
     /// @brief Returns `true` if the parameter associated with `key` is a
     /// string; `false` otherwise.
@@ -926,6 +1157,8 @@ namespace awkward {
 
     /// @brief Concatenates this array with `other` by creating a
     /// {@link UnionArrayOf UnionArray} instead of actually merging the data.
+    /// This does not call
+    /// {@link UnionArrayOf#simplify_uniontype UnionArray::simplify_uniontype}.
     const ContentPtr
       merge_as_union(const ContentPtr& other) const;
 
@@ -1086,6 +1319,32 @@ namespace awkward {
                           const SliceJagged64& slicecontent,
                           const Slice& tail) const = 0;
 
+    /// @brief Internal function that propagates a jagged array (array with
+    /// irregular-length dimensions) slice through one axis.
+    ///
+    /// See generic #getitem_next_jagged for details.
+    virtual const ContentPtr
+      getitem_next_jagged(const Index64& slicestarts,
+                          const Index64& slicestops,
+                          const SliceVarNewAxis& slicecontent,
+                          const Slice& tail) const = 0;
+
+    /// @brief Internal function that propagates a generic #getitem request
+    /// through one axis (including advanced indexing).
+    ///
+    /// See generic #getitem_next for details.
+    virtual const ContentPtr
+      getitem_next(const SliceVarNewAxis& varnewaxis,
+                   const Slice& tail,
+                   const Index64& advanced) const = 0;
+
+    /// @brief Internal function to convert a SliceVarNewAxis into the
+    /// right-sized SliceJagged64 for a given Content.
+    ///
+    /// See generic #getitem_next_jagged for details.
+    virtual const SliceJagged64
+      varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const = 0;
+
     /// @brief Internal function defining the negative axis handling for many
     /// operations.
     ///
@@ -1100,13 +1359,38 @@ namespace awkward {
     /// This allows a RecordArray or {@link UnionArrayOf UnionArray} with
     /// different depths to accept `axis = -1` as the last axis, regardless
     /// of how deep that is in different record fields or union possibilities.
+    const int64_t
+      axis_wrap_if_negative(int64_t axis) const;
+
+    /// @brief Recursively copies components of the array from main memory to a
+    /// GPU (if `ptr_lib == kernel::lib::cuda`) or to main memory (if
+    /// `ptr_lib == kernel::lib::cpu`) if those components are not already there.
+    virtual const ContentPtr
+      copy_to(kernel::lib ptr_lib) const = 0;
+
+    /// @brief Change the leaf types to 'totype'.
+    virtual const ContentPtr
+      numbers_to_type(const std::string& name) const = 0;
+
+    /// @brief Returns 'true' if all components of the array are unique
+    virtual bool
+      is_unique() const = 0;
+
+    /// @brief Returns sorted unique values.
     ///
-    /// @note This function has not been implemented:
-    /// [scikit-hep/awkward-1.0#163](https://github.com/scikit-hep/awkward-1.0/issues/163).
-    static int64_t
-      axis_wrap_if_negative(int64_t axis);
+    virtual const ContentPtr
+      unique() const = 0;
+
+    /// @brief Returns 'true' if subranges are equal
+    virtual bool
+      is_subrange_equal(const Index64& start, const Index64& stop) const = 0;
 
   protected:
+    /// @brief Internal function to determine if a particular set of parameters
+    /// are allowed on a particular node (centralized by parameter, not by class).
+    const std::string
+      validityerror_parameters(const std::string& path) const;
+
     /// @brief Internal function to wrap putative #getitem output with enough
     /// RegularArray nodes to satisfy a given `shape`.
     ///
@@ -1128,6 +1412,10 @@ namespace awkward {
                           const std::string& pre,
                           const std::string& post) const;
 
+    /// @brief Internal function to support most of the logic in #kernels.
+    kernel::lib
+      kernels_compare(kernel::lib from_index, const ContentPtr& content) const;
+
   protected:
     /// @brief See #identities.
     IdentitiesPtr identities_;
