@@ -1,4 +1,4 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_SLICE_H_
 #define AWKWARD_SLICE_H_
@@ -7,8 +7,7 @@
 #include <vector>
 #include <memory>
 
-#include "awkward/cpu-kernels/getitem.h"
-#include "awkward/type/Type.h"
+#include "awkward/common.h"
 #include "awkward/Index.h"
 
 namespace awkward {
@@ -19,7 +18,7 @@ namespace awkward {
   ///
   /// @brief Abstract class for slice items, which are elements of a tuple
   /// passed to an array's `__getitem__` in Python.
-  class EXPORT_SYMBOL SliceItem {
+  class LIBAWKWARD_EXPORT_SYMBOL SliceItem {
   public:
     /// @brief Virtual destructor acts as a first non-inline virtual function
     /// that determines a specific translation unit in which vtable shall be
@@ -30,6 +29,11 @@ namespace awkward {
     virtual const SliceItemPtr
       shallow_copy() const = 0;
 
+    /// @brief For SliceVarNewAxis to spread the slice in a way similar to
+    /// Content::carry.
+    virtual const SliceItemPtr
+      carry(const Index64& carry) const;
+
     /// @brief Returns a string representation of this slice item (single-line
     /// custom format).
     virtual const std::string
@@ -43,13 +47,20 @@ namespace awkward {
     /// {@link Content#getitem_next Content::getitem_next}.
     virtual bool
       preserves_type(const Index64& advanced) const = 0;
+
+    /// @brief Returns `true` if this slice item has all the same buffers and
+    /// parameters as `other`; `false` otherwise.
+    ///
+    /// @param other The slice item to compare this with.
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const = 0;
   };
 
   /// @class SliceAt
   ///
   /// @brief Represents an integer in a tuple of slices passed to
   /// `__getitem__` in Python.
-  class EXPORT_SYMBOL SliceAt: public SliceItem {
+  class LIBAWKWARD_EXPORT_SYMBOL SliceAt: public SliceItem {
   public:
     /// @brief Creates a SliceAt from a full set of parameters.
     ///
@@ -72,6 +83,9 @@ namespace awkward {
     bool
       preserves_type(const Index64& advanced) const override;
 
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
   private:
     /// @brief See #at.
     const int64_t at_;
@@ -81,7 +95,7 @@ namespace awkward {
   ///
   /// @brief Represents a Python `slice` object (usual syntax:
   /// `array[start:stop:step]`).
-  class EXPORT_SYMBOL SliceRange: public SliceItem {
+  class LIBAWKWARD_EXPORT_SYMBOL SliceRange: public SliceItem {
   public:
     /// @brief Creates a SliceRange from a full set of parameters.
     ///
@@ -141,6 +155,9 @@ namespace awkward {
     bool
       preserves_type(const Index64& advanced) const override;
 
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
   private:
     /// @brief See #start.
     const int64_t start_;
@@ -154,7 +171,7 @@ namespace awkward {
   ///
   /// @brief Represents a Python `Ellipsis` object (usual syntax:
   /// `array[...]`).
-  class EXPORT_SYMBOL SliceEllipsis: public SliceItem {
+  class LIBAWKWARD_EXPORT_SYMBOL SliceEllipsis: public SliceItem {
   public:
     /// @brief Creates a SliceEllipsis.
     SliceEllipsis();
@@ -170,6 +187,9 @@ namespace awkward {
     /// Always `true` for SliceEllipsis.
     bool
       preserves_type(const Index64& advanced) const override;
+
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
   };
 
   /// @class SliceNewAxis
@@ -179,7 +199,7 @@ namespace awkward {
   /// marker (a.k.a. `None`), which prompts `__getitem__` to insert a
   /// length-1 regular dimension (RegularArray) at some point in the slice
   /// tuple.
-  class EXPORT_SYMBOL SliceNewAxis: public SliceItem {
+  class LIBAWKWARD_EXPORT_SYMBOL SliceNewAxis: public SliceItem {
   public:
     /// @brief Creates a SliceNewAxis.
     SliceNewAxis();
@@ -195,6 +215,9 @@ namespace awkward {
     /// Always `false` for SliceNewAxis.
     bool
       preserves_type(const Index64& advanced) const override;
+
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
   };
 
   /// @class SliceArrayOf
@@ -204,7 +227,11 @@ namespace awkward {
   ///
   /// Currently, the only type specialization is `T = int64_t`.
   template <typename T>
-  class EXPORT_SYMBOL SliceArrayOf: public SliceItem {
+  class
+#ifdef AWKWARD_SLICE_NO_EXTERN_TEMPLATE
+  LIBAWKWARD_EXPORT_SYMBOL
+#endif
+  SliceArrayOf: public SliceItem {
   public:
     /// @brief Creates a SliceArrayOf from a full set of parameters.
     ///
@@ -273,6 +300,9 @@ namespace awkward {
     const SliceItemPtr
       shallow_copy() const override;
 
+    const SliceItemPtr
+      carry(const Index64& carry) const override;
+
     const std::string
       tostring() const override;
 
@@ -286,6 +316,9 @@ namespace awkward {
     bool
       preserves_type(const Index64& advanced) const override;
 
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
     /// @brief Returns a one-dimensional contiguous version of the array,
     /// like NumPy's [ravel](https://docs.scipy.org/doc/numpy/reference/generated/numpy.ravel.html).
     const IndexOf<T>
@@ -302,7 +335,7 @@ namespace awkward {
     bool frombool_;
   };
 
-#if !defined AWKWARD_SLICE_NO_EXTERN_TEMPLATE && !defined _MSC_VER
+#ifndef AWKWARD_SLICE_NO_EXTERN_TEMPLATE
   extern template class SliceArrayOf<int64_t>;
 #endif
 
@@ -312,7 +345,7 @@ namespace awkward {
   ///
   /// @brief Represents a single string in a slice tuple, indicating that a
   /// RecordArray should be replaced by one of its fields.
-  class EXPORT_SYMBOL SliceField: public SliceItem {
+  class LIBAWKWARD_EXPORT_SYMBOL SliceField: public SliceItem {
   public:
     /// @brief Creates a SliceField from a full set of parameters.
     ///
@@ -344,6 +377,9 @@ namespace awkward {
     bool
       preserves_type(const Index64& advanced) const override;
 
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
   private:
     /// @brief See #key.
     const std::string key_;
@@ -353,7 +389,7 @@ namespace awkward {
   ///
   /// @brief Represents a list of strings in a slice tuple, indicating that a
   /// RecordArray should be replaced by a subset of its fields.
-  class EXPORT_SYMBOL SliceFields: public SliceItem {
+  class LIBAWKWARD_EXPORT_SYMBOL SliceFields: public SliceItem {
   public:
     /// @brief Creates a SliceFields from a full set of parameters.
     ///
@@ -385,6 +421,9 @@ namespace awkward {
     bool
       preserves_type(const Index64& advanced) const override;
 
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
   private:
     /// @brief See #keys.
     const std::vector<std::string> keys_;
@@ -397,7 +436,11 @@ namespace awkward {
   ///
   /// Currently, the only type specialization is `T = int64_t`.
   template <typename T>
-  class EXPORT_SYMBOL SliceMissingOf: public SliceItem {
+  class
+#ifdef AWKWARD_SLICE_NO_EXTERN_TEMPLATE
+  LIBAWKWARD_EXPORT_SYMBOL
+#endif
+  SliceMissingOf: public SliceItem {
   public:
     /// @brief Creates a SliceMissingOf with a full set of parameters.
     ///
@@ -433,6 +476,9 @@ namespace awkward {
     const SliceItemPtr
       shallow_copy() const override;
 
+    const SliceItemPtr
+      carry(const Index64& carry) const override;
+
     const std::string
       tostring() const override;
 
@@ -447,6 +493,9 @@ namespace awkward {
     bool
       preserves_type(const Index64& advanced) const override;
 
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
   private:
     /// @brief See #index.
     const IndexOf<T> index_;
@@ -456,7 +505,7 @@ namespace awkward {
     const SliceItemPtr content_;
   };
 
-#if !defined AWKWARD_SLICE_NO_EXTERN_TEMPLATE && !defined _MSC_VER
+#ifndef AWKWARD_SLICE_NO_EXTERN_TEMPLATE
   extern template class SliceMissingOf<int64_t>;
 #endif
 
@@ -470,7 +519,11 @@ namespace awkward {
   ///
   /// Currently, the only type specialization is `T = int64_t`.
   template <typename T>
-  class EXPORT_SYMBOL SliceJaggedOf: public SliceItem {
+  class
+#ifdef AWKWARD_SLICE_NO_EXTERN_TEMPLATE
+  LIBAWKWARD_EXPORT_SYMBOL
+#endif
+  SliceJaggedOf: public SliceItem {
   public:
     /// @brief Creates a SliceJaggedOf with a full set of parameters.
     ///
@@ -506,6 +559,9 @@ namespace awkward {
     const SliceItemPtr
       shallow_copy() const override;
 
+    const SliceItemPtr
+      carry(const Index64& carry) const override;
+
     const std::string
       tostring() const override;
 
@@ -520,6 +576,9 @@ namespace awkward {
     bool
       preserves_type(const Index64& advanced) const override;
 
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
   private:
     /// @brief See #offsets.
     const IndexOf<T> offsets_;
@@ -527,7 +586,7 @@ namespace awkward {
     const SliceItemPtr content_;
   };
 
-#if !defined AWKWARD_SLICE_NO_EXTERN_TEMPLATE && !defined _MSC_VER
+#ifndef AWKWARD_SLICE_NO_EXTERN_TEMPLATE
   extern template class SliceJaggedOf<int64_t>;
 #endif
 
@@ -537,7 +596,7 @@ namespace awkward {
   ///
   /// @brief A sequence of SliceItem objects representing a tuple passed
   /// to Python's `__getitem__`.
-  class EXPORT_SYMBOL Slice {
+  class LIBAWKWARD_EXPORT_SYMBOL Slice {
   public:
     /// @brief Represents a missing {@link SliceRange#start start},
     /// {@link SliceRange#stop stop}, or {@link SliceRange#step step}
@@ -585,6 +644,14 @@ namespace awkward {
     const Slice
       tail() const;
 
+    /// @brief Returns a Slice containing only SliceField and SliceFields.
+    const Slice
+      only_fields() const;
+
+    /// @brief Returns a Slice without any SliceField or SliceFields.
+    const Slice
+      not_fields() const;
+
     /// @brief Returns a string representation of this slice item (single-line
     /// custom format).
     const std::string
@@ -649,6 +716,13 @@ namespace awkward {
     bool
       isadvanced() const;
 
+    /// @brief Returns `true` if this slice has all the same buffers and
+    /// parameters as `other`; `false` otherwise.
+    ///
+    /// @param other The slice to compare this with.
+    bool
+      referentially_equal(const Slice& other) const;
+
   private:
     /// @brief See #items.
     std::vector<SliceItemPtr> items_;
@@ -656,6 +730,31 @@ namespace awkward {
     bool sealed_;
   };
 
+  class LIBAWKWARD_EXPORT_SYMBOL SliceVarNewAxis: public SliceItem {
+  public:
+    SliceVarNewAxis(const SliceItemPtr& content);
+
+    const SliceItemPtr
+      content() const;
+
+    const SliceItemPtr
+      shallow_copy() const override;
+
+    const SliceItemPtr
+      carry(const Index64& carry) const override;
+
+    const std::string
+      tostring() const override;
+
+    bool
+      preserves_type(const Index64& advanced) const override;
+
+    virtual bool
+      referentially_equal(const SliceItemPtr& other) const override;
+
+  private:
+    const SliceItemPtr content_;
+  };
 }
 
 #endif // AWKWARD_SLICE_H_
