@@ -1,4 +1,4 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_EMPTYARRAY_H_
 #define AWKWARD_EMPTYARRAY_H_
@@ -7,7 +7,7 @@
 #include <memory>
 #include <vector>
 
-#include "awkward/cpu-kernels/util.h"
+#include "awkward/common.h"
 #include "awkward/Slice.h"
 #include "awkward/Content.h"
 
@@ -17,11 +17,12 @@ namespace awkward {
   /// @class EmptyForm
   ///
   /// @brief Form describing EmptyArray.
-  class EXPORT_SYMBOL EmptyForm: public Form {
+  class LIBAWKWARD_EXPORT_SYMBOL EmptyForm: public Form {
   public:
     /// @brief Creates a EmptyForm. See EmptyArray for documentation.
     EmptyForm(bool has_identities,
-              const util::Parameters& parameters);
+              const util::Parameters& parameters,
+              const FormKey& form_key);
 
     const TypePtr
       type(const util::TypeStrs& typestrs) const override;
@@ -32,6 +33,9 @@ namespace awkward {
     const FormPtr
       shallow_copy() const override;
 
+    const FormPtr
+      with_form_key(const FormKey& form_key) const override;
+
     const std::string
       purelist_parameter(const std::string& key) const override;
 
@@ -41,6 +45,9 @@ namespace awkward {
     int64_t
       purelist_depth() const override;
 
+    bool
+      dimension_optiontype() const override;
+
     const std::pair<int64_t, int64_t>
       minmax_depth() const override;
 
@@ -65,7 +72,16 @@ namespace awkward {
     bool
       equal(const FormPtr& other,
             bool check_identities,
-            bool check_parameters) const override;
+            bool check_parameters,
+            bool check_form_key,
+            bool compatibility_check) const override;
+
+    const FormPtr
+      getitem_field(const std::string& key) const override;
+
+    const FormPtr
+      getitem_fields(const std::vector<std::string>& keys) const override;
+
   };
 
   /// @class EmptyArray
@@ -74,7 +90,7 @@ namespace awkward {
   /// consequence) UnknownType.
   ///
   /// See #EmptyArray for the meaning of each parameter.
-  class EXPORT_SYMBOL EmptyArray: public Content {
+  class LIBAWKWARD_EXPORT_SYMBOL EmptyArray: public Content {
   public:
     /// @brief Creates an EmptyArray from a full set of parameters.
     ///
@@ -89,7 +105,8 @@ namespace awkward {
     /// {NumpyArray#format format}.
     const std::shared_ptr<NumpyArray>
       toNumpyArray(const std::string& format,
-                   ssize_t itemsize) const;
+                   ssize_t itemsize,
+                   util::dtype npdtype) const;
 
     /// @brief User-friendly name of this class: `"EmptyArray"`.
     const std::string
@@ -107,11 +124,11 @@ namespace awkward {
     const FormPtr
       form(bool materialize) const override;
 
-    bool
-      has_virtual_form() const override;
+    kernel::lib
+      kernels() const override;
 
-    bool
-      has_virtual_length() const override;
+    void
+      caches(std::vector<ArrayCachePtr>& out) const override;
 
     const std::string
       tostring_part(const std::string& indent,
@@ -156,11 +173,28 @@ namespace awkward {
     const ContentPtr
       getitem_field(const std::string& key) const override;
 
+    const ContentPtr
+      getitem_field(const std::string& key,
+                    const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_fields(const std::vector<std::string>& keys) const override;
 
     const ContentPtr
-      carry(const Index64& carry) const override;
+      getitem_fields(const std::vector<std::string>& keys,
+                     const Slice& only_fields) const override;
+
+    const ContentPtr
+      carry(const Index64& carry, bool allow_lazy) const override;
+
+    int64_t
+      purelist_depth() const override;
+
+    const std::pair<int64_t, int64_t>
+      minmax_depth() const override;
+
+    const std::pair<bool, int64_t>
+      branch_depth() const override;
 
     int64_t
       numfields() const override;
@@ -196,8 +230,11 @@ namespace awkward {
     bool
       mergeable(const ContentPtr& other, bool mergebool) const override;
 
+    bool
+      referentially_equal(const ContentPtr& other) const override;
+
     const ContentPtr
-      merge(const ContentPtr& other) const override;
+      mergemany(const ContentPtrVec& others) const override;
 
     const SliceItemPtr
       asslice() const override;
@@ -217,11 +254,30 @@ namespace awkward {
       reduce_next(const Reducer& reducer,
                   int64_t negaxis,
                   const Index64& starts,
+                  const Index64& shifts,
                   const Index64& parents,
                   int64_t outlength,
                   bool mask,
                   bool keepdims) const override;
 
+    const ContentPtr
+      sort_next(int64_t negaxis,
+                const Index64& starts,
+                const Index64& parents,
+                int64_t outlength,
+                bool ascending,
+                bool stable,
+                bool keepdims) const override;
+
+    const ContentPtr
+      argsort_next(int64_t negaxis,
+                   const Index64& starts,
+                   const Index64& parents,
+                   int64_t outlength,
+                   bool ascending,
+                   bool stable,
+                   bool keepdims) const override;
+
     const ContentPtr
       localindex(int64_t axis, int64_t depth) const override;
 
@@ -280,6 +336,39 @@ namespace awkward {
                           const Index64& slicestops,
                           const SliceJagged64& slicecontent,
                           const Slice& tail) const override;
+
+    const ContentPtr
+      getitem_next_jagged(const Index64& slicestarts,
+                          const Index64& slicestops,
+                          const SliceVarNewAxis& slicecontent,
+                          const Slice& tail) const override;
+
+    const ContentPtr
+      getitem_next(const SliceVarNewAxis& varnewaxis,
+                   const Slice& tail,
+                   const Index64& advanced) const override;
+
+    const SliceJagged64
+      varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const override;
+
+    const ContentPtr
+      copy_to(kernel::lib ptr_lib) const override;
+
+    const ContentPtr
+      numbers_to_type(const std::string& name) const override;
+
+    /// @brief Returns 'true' if all components of the array are unique
+    bool
+      is_unique() const override;
+
+    /// @brief Returns an array where all components are unique
+    const ContentPtr
+      unique() const override;
+
+    /// @brief Returns 'true' if subranges are equal
+    bool
+      is_subrange_equal(const Index64& start, const Index64& stop) const override;
+
   };
 
 }
