@@ -1,11 +1,11 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_LISTOFFSETARRAY_H_
 #define AWKWARD_LISTOFFSETARRAY_H_
 
 #include <memory>
 
-#include "awkward/cpu-kernels/util.h"
+#include "awkward/common.h"
 #include "awkward/Index.h"
 #include "awkward/Identities.h"
 #include "awkward/Content.h"
@@ -14,12 +14,13 @@ namespace awkward {
   /// @class ListOffsetForm
   ///
   /// @brief Form describing ListOffsetArray.
-  class EXPORT_SYMBOL ListOffsetForm: public Form {
+  class LIBAWKWARD_EXPORT_SYMBOL ListOffsetForm: public Form {
   public:
     /// @brief Creates a ListOffsetForm. See
     /// {@link ListOffsetArrayOf ListOffsetArray} for documentation.
     ListOffsetForm(bool has_identities,
                    const util::Parameters& parameters,
+                   const FormKey& form_key,
                    Index::Form offsets,
                    const FormPtr& content);
 
@@ -38,6 +39,9 @@ namespace awkward {
     const FormPtr
       shallow_copy() const override;
 
+    const FormPtr
+      with_form_key(const FormKey& form_key) const override;
+
     const std::string
       purelist_parameter(const std::string& key) const override;
 
@@ -47,6 +51,9 @@ namespace awkward {
     int64_t
       purelist_depth() const override;
 
+    bool
+      dimension_optiontype() const override;
+
     const std::pair<int64_t, int64_t>
       minmax_depth() const override;
 
@@ -71,7 +78,15 @@ namespace awkward {
     bool
       equal(const FormPtr& other,
             bool check_identities,
-            bool check_parameters) const override;
+            bool check_parameters,
+            bool check_form_key,
+            bool compatibility_check) const override;
+
+    const FormPtr
+      getitem_field(const std::string& key) const override;
+
+    const FormPtr
+      getitem_fields(const std::vector<std::string>& keys) const override;
 
   private:
     Index::Form offsets_;
@@ -90,7 +105,11 @@ namespace awkward {
   ///
   /// See #ListOffsetArrayOf for the meaning of each parameter.
   template <typename T>
-  class EXPORT_SYMBOL ListOffsetArrayOf: public Content {
+  class
+#ifdef AWKWARD_LISTOFFSETARRAY_NO_EXTERN_TEMPLATE
+  LIBAWKWARD_EXPORT_SYMBOL
+#endif
+  ListOffsetArrayOf: public Content {
   public:
     /// @brief Creates a ListOffsetArray from a full set of parameters.
     ///
@@ -110,7 +129,8 @@ namespace awkward {
     ListOffsetArrayOf<T>(const IdentitiesPtr& identities,
                          const util::Parameters& parameters,
                          const IndexOf<T>& offsets,
-                         const ContentPtr& content);
+                         const ContentPtr& content,
+                         bool represents_regular = false);
 
     /// @brief Positions where one nested list stops and the next starts in
     /// the #content; the `offsets` must be monotonically increasing.
@@ -195,11 +215,11 @@ namespace awkward {
     const FormPtr
       form(bool materialize) const override;
 
-    bool
-      has_virtual_form() const override;
+    kernel::lib
+      kernels() const override;
 
-    bool
-      has_virtual_length() const override;
+    void
+      caches(std::vector<ArrayCachePtr>& out) const override;
 
     const std::string
       tostring_part(const std::string& indent,
@@ -247,9 +267,17 @@ namespace awkward {
     const ContentPtr
       getitem_field(const std::string& key) const override;
 
+    const ContentPtr
+      getitem_field(const std::string& key,
+                    const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_fields(const std::vector<std::string>& keys) const override;
 
+    const ContentPtr
+      getitem_fields(const std::vector<std::string>& keys,
+                     const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_next_jagged(const Index64& slicestarts,
                           const Index64& slicestops,
@@ -257,7 +285,16 @@ namespace awkward {
                           const Slice& tail) const override;
 
     const ContentPtr
-      carry(const Index64& carry) const override;
+      carry(const Index64& carry, bool allow_lazy) const override;
+
+    int64_t
+      purelist_depth() const override;
+
+    const std::pair<int64_t, int64_t>
+      minmax_depth() const override;
+
+    const std::pair<bool, int64_t>
+      branch_depth() const override;
 
     int64_t
       numfields() const override;
@@ -294,8 +331,11 @@ namespace awkward {
     bool
       mergeable(const ContentPtr& other, bool mergebool) const override;
 
+    bool
+      referentially_equal(const ContentPtr& other) const override;
+
     const ContentPtr
-      merge(const ContentPtr& other) const override;
+      mergemany(const ContentPtrVec& others) const override;
 
     const SliceItemPtr
       asslice() const override;
@@ -315,11 +355,30 @@ namespace awkward {
       reduce_next(const Reducer& reducer,
                   int64_t negaxis,
                   const Index64& starts,
+                  const Index64& shifts,
                   const Index64& parents,
                   int64_t outlength,
                   bool mask,
                   bool keepdims) const override;
 
+    const ContentPtr
+      sort_next(int64_t negaxis,
+                const Index64& starts,
+                const Index64& parents,
+                int64_t outlength,
+                bool ascending,
+                bool stable,
+                bool keepdims) const override;
+
+    const ContentPtr
+      argsort_next(int64_t negaxis,
+                   const Index64& starts,
+                   const Index64& parents,
+                   int64_t outlength,
+                   bool ascending,
+                   bool stable,
+                   bool keepdims) const override;
+
     const ContentPtr
       localindex(int64_t axis, int64_t depth) const override;
 
@@ -369,14 +428,47 @@ namespace awkward {
                           const SliceJagged64& slicecontent,
                           const Slice& tail) const override;
 
+    const ContentPtr
+      getitem_next_jagged(const Index64& slicestarts,
+                          const Index64& slicestops,
+                          const SliceVarNewAxis& slicecontent,
+                          const Slice& tail) const override;
+
+    const ContentPtr
+      getitem_next(const SliceVarNewAxis& varnewaxis,
+                   const Slice& tail,
+                   const Index64& advanced) const override;
+
+    const SliceJagged64
+      varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const override;
+
+    const ContentPtr
+      copy_to(kernel::lib ptr_lib) const override;
+
+    const ContentPtr
+      numbers_to_type(const std::string& name) const override;
+
+    /// @brief Returns 'true' if all components of the array are unique
+    bool
+      is_unique() const override;
+
+    /// @brief Returns an array where all components are unique
+    const ContentPtr
+      unique() const override;
+
+    /// @brief Returns 'true' if subranges are equal
+    bool
+      is_subrange_equal(const Index64& start, const Index64& stop) const override;
+
   private:
     /// @brief See #offsets.
     const IndexOf<T> offsets_;
     /// @brief See #content.
     const ContentPtr content_;
+    const bool represents_regular_;
   };
 
-#if !defined AWKWARD_LISTOFFSETARRAY_NO_EXTERN_TEMPLATE && !defined _MSC_VER
+#ifndef AWKWARD_LISTOFFSETARRAY_NO_EXTERN_TEMPLATE
   extern template class ListOffsetArrayOf<int32_t>;
   extern template class ListOffsetArrayOf<uint32_t>;
   extern template class ListOffsetArrayOf<int64_t>;
