@@ -1,15 +1,13 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_NUMPYARRAY_H_
 #define AWKWARD_NUMPYARRAY_H_
 
 #include <string>
-#include <unordered_map>
 #include <memory>
-#include <typeindex>
 #include <vector>
 
-#include "awkward/cpu-kernels/util.h"
+#include "awkward/common.h"
 #include "awkward/Slice.h"
 #include "awkward/Content.h"
 
@@ -17,14 +15,16 @@ namespace awkward {
   /// @class NumpyForm
   ///
   /// @brief Form describing NumpyArray.
-  class EXPORT_SYMBOL NumpyForm: public Form {
+  class LIBAWKWARD_EXPORT_SYMBOL NumpyForm: public Form {
   public:
     /// @brief Creates a NumpyForm. See NumpyArray for documentation.
     NumpyForm(bool has_identities,
               const util::Parameters& parameters,
+              const FormKey& form_key,
               const std::vector<int64_t>& inner_shape,
               int64_t itemsize,
-              const std::string& format);
+              const std::string& format,
+              util::dtype dtype);
 
     const std::vector<int64_t>
       inner_shape() const;
@@ -32,7 +32,11 @@ namespace awkward {
     int64_t
       itemsize() const;
 
-    const std::string format() const;
+    const std::string
+      format() const;
+
+    util::dtype
+      dtype() const;
 
     const std::string
       primitive() const;
@@ -55,6 +59,9 @@ namespace awkward {
     const FormPtr
       shallow_copy() const override;
 
+    const FormPtr
+      with_form_key(const FormKey& form_key) const override;
+
     const std::string
       purelist_parameter(const std::string& key) const override;
 
@@ -64,6 +71,9 @@ namespace awkward {
     int64_t
       purelist_depth() const override;
 
+    bool
+      dimension_optiontype() const override;
+
     const std::pair<int64_t, int64_t>
       minmax_depth() const override;
 
@@ -88,12 +98,21 @@ namespace awkward {
     bool
       equal(const FormPtr& other,
             bool check_identities,
-            bool check_parameters) const override;
+            bool check_parameters,
+            bool check_form_key,
+            bool compatibility_check) const override;
+
+    const FormPtr
+      getitem_field(const std::string& key) const override;
+
+    const FormPtr
+      getitem_fields(const std::vector<std::string>& keys) const override;
 
   private:
     const std::vector<int64_t> inner_shape_;
     int64_t itemsize_;
     const std::string format_;
+    const util::dtype dtype_;
   };
 
   /// @class NumpyArray
@@ -113,7 +132,7 @@ namespace awkward {
   ///  - 8-bit booleans
   ///
   ///(native endian only).
-  class EXPORT_SYMBOL NumpyArray: public Content {
+  class LIBAWKWARD_EXPORT_SYMBOL NumpyArray: public Content {
   public:
 
     /// @brief Creates a NumpyArray from a full set of parameters.
@@ -144,6 +163,9 @@ namespace awkward {
     /// pybind11). Note that 32-bit systems and Windows use "`q/Q`" for
     /// signed/unsigned 64-bit and "`l/L`" for 32-bit, while all other systems
     /// use "`l/L`" for 64-bit and "`i/I`" for 32-bit.
+    /// @param dtype Enumeration for the type without the platform-dependence
+    /// that `format` has.
+    /// @param ptr_lib Indicates the kernel libraries to use for this `ptr`.
     ///
     /// Note that the integer type for all of these parameters is `ssize_t`,
     /// for consistency with pybind11. Most integers in Awkward are `int64_t`.
@@ -154,38 +176,33 @@ namespace awkward {
                const std::vector<ssize_t>& strides,
                ssize_t byteoffset,
                ssize_t itemsize,
-               const std::string format);
+               const std::string& format,
+               util::dtype dtype,
+               const kernel::lib ptr_lib);
 
     /// @brief Creates a NumpyArray from an {@link IndexOf Index8}.
-    NumpyArray(const Index8 index);
+    explicit NumpyArray(const Index8 index);
     /// @brief Creates a NumpyArray from an {@link IndexOf IndexU8}.
-    NumpyArray(const IndexU8 index);
+    explicit NumpyArray(const IndexU8 index);
     /// @brief Creates a NumpyArray from an {@link IndexOf Index32}.
-    NumpyArray(const Index32 index);
+    explicit NumpyArray(const Index32 index);
     /// @brief Creates a NumpyArray from an {@link IndexOf IndexU32}.
-    NumpyArray(const IndexU32 index);
+    explicit NumpyArray(const IndexU32 index);
     /// @brief Creates a NumpyArray from an {@link IndexOf Index64}.
-    NumpyArray(const Index64 index);
-    /// @brief Creates a NumpyArray from an {@link IndexOf Index8}.
-    /// The format may be specified explicitly.
-    NumpyArray(const Index8 index, const std::string& format);
-    /// @brief Creates a NumpyArray from an {@link IndexOf IndexU8}.
-    /// The format may be specified explicitly.
-    NumpyArray(const IndexU8 index, const std::string& format);
-    /// @brief Creates a NumpyArray from an {@link IndexOf Index32}.
-    /// The format may be specified explicitly.
-    NumpyArray(const Index32 index, const std::string& format);
-    /// @brief Creates a NumpyArray from an {@link IndexOf IndexU32}.
-    /// The format may be specified explicitly.
-    NumpyArray(const IndexU32 index, const std::string& format);
-    /// @brief Creates a NumpyArray from an {@link IndexOf Index64}.
-    /// The format may be specified explicitly.
-    NumpyArray(const Index64 index, const std::string& format);
+    explicit NumpyArray(const Index64 index);
 
     /// @brief Reference-counted pointer to the array buffer.
     const std::shared_ptr<void>
       ptr() const;
 
+    /// @param ptr_lib Indicates the kernel libraries to use for this `ptr`.
+    kernel::lib
+      ptr_lib() const;
+
+    /// @brief Raw pointer to the beginning of data (i.e. offset accounted for).
+    void*
+      data() const;
+
     /// @brief Number of elements in each dimension. A one-dimensional
     /// array has a shape of length one.
     ///
@@ -242,6 +259,11 @@ namespace awkward {
     const std::string
       format() const;
 
+    /// @param dtype Enumeration for the type without the platform-dependence
+    /// that `format` has.
+    util::dtype
+      dtype() const;
+
     /// @brief The number of dimensions, which is `shape.size()`.
     ///
     /// Note that the integer type is `ssize_t` for consistency with pybind11.
@@ -256,17 +278,6 @@ namespace awkward {
     bool
       isempty() const;
 
-    /// @brief An untyped pointer to item zero in the buffer.
-    void*
-      byteptr() const;
-
-    /// @brief An untyped pointer to item `at` in the buffer.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    void*
-      byteptr(ssize_t at) const;
-
     /// @brief The length of the array (or scalar, if `shape.empty()`) in
     /// bytes.
     ///
@@ -282,76 +293,6 @@ namespace awkward {
     uint8_t
       getbyte(ssize_t at) const;
 
-    /// @brief Dereferences a selected item as a `int8_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    int8_t
-      getint8(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `uint8_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    uint8_t
-      getuint8(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `int16_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    int16_t
-      getint16(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `uint16_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    uint16_t
-      getuint16(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `int32_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    int32_t
-      getint32(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `uint32_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    uint32_t
-      getuint32(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `int64_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    int64_t
-      getint64(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `uint64_t`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    uint64_t
-      getuint64(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `float`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    float_t
-      getfloat(ssize_t at) const;
-
-    /// @brief Dereferences a selected item as a `double`.
-    ///
-    /// Note that the integer type is `ssize_t` for consistency with pybind11.
-    /// Most integers in Awkward are `int64_t`.
-    double_t
-      getdouble(ssize_t at) const;
-
     /// @brief A contiguous version of this array with multidimensional
     /// #shape replaced by nested RegularArray nodes.
     ///
@@ -380,11 +321,11 @@ namespace awkward {
     const FormPtr
       form(bool materialize) const override;
 
-    bool
-      has_virtual_form() const override;
+    kernel::lib
+      kernels() const override;
 
-    bool
-      has_virtual_length() const override;
+    void
+      caches(std::vector<ArrayCachePtr>& out) const override;
 
     const std::string
       tostring_part(const std::string& indent,
@@ -429,9 +370,17 @@ namespace awkward {
     const ContentPtr
       getitem_field(const std::string& key) const override;
 
+    const ContentPtr
+      getitem_field(const std::string& key,
+                    const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_fields(const std::vector<std::string>& keys) const override;
 
+    const ContentPtr
+      getitem_fields(const std::vector<std::string>& keys,
+                     const Slice& only_fields) const override;
+
     const ContentPtr
       getitem(const Slice& where) const override;
 
@@ -441,7 +390,16 @@ namespace awkward {
                    const Index64& advanced) const override;
 
     const ContentPtr
-      carry(const Index64& carry) const override;
+      carry(const Index64& carry, bool allow_lazy) const override;
+
+    int64_t
+      purelist_depth() const override;
+
+    const std::pair<int64_t, int64_t>
+      minmax_depth() const override;
+
+    const std::pair<bool, int64_t>
+      branch_depth() const override;
 
     int64_t
       numfields() const override;
@@ -477,8 +435,11 @@ namespace awkward {
     bool
       mergeable(const ContentPtr& other, bool mergebool) const override;
 
+    bool
+      referentially_equal(const ContentPtr& other) const override;
+
     const ContentPtr
-      merge(const ContentPtr& other) const override;
+      mergemany(const ContentPtrVec& others) const override;
 
     const SliceItemPtr
       asslice() const override;
@@ -498,11 +459,37 @@ namespace awkward {
       reduce_next(const Reducer& reducer,
                   int64_t negaxis,
                   const Index64& starts,
+                  const Index64& shifts,
                   const Index64& parents,
                   int64_t outlength,
                   bool mask,
                   bool keepdims) const override;
 
+    const ContentPtr
+      sort_next(int64_t negaxis,
+                const Index64& starts,
+                const Index64& parents,
+                int64_t outlength,
+                bool ascending,
+                bool stable,
+                bool keepdims) const override;
+
+    const ContentPtr
+      sort_asstrings(const Index64& offsets,
+                     bool ascending,
+                     bool stable) const;
+    const ContentPtr
+      as_unique_strings(const Index64& offsets) const;
+
+    const ContentPtr
+      argsort_next(int64_t negaxis,
+                   const Index64& starts,
+                   const Index64& parents,
+                   int64_t outlength,
+                   bool ascending,
+                   bool stable,
+                   bool keepdims) const override;
+
     const ContentPtr
       localindex(int64_t axis, int64_t depth) const override;
 
@@ -587,6 +574,34 @@ namespace awkward {
                    const Slice& tail,
                    const Index64& advanced) const override;
 
+    const ContentPtr
+      getitem_next(const SliceVarNewAxis& varnewaxis,
+                   const Slice& tail,
+                   const Index64& advanced) const override;
+
+    const SliceJagged64
+      varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const override;
+
+    /// @brief An utility function to create a new instance of NumpyArray on the
+    /// GPU identical to this one.
+    const ContentPtr
+      copy_to(kernel::lib ptr_lib) const override;
+
+    const ContentPtr
+      numbers_to_type(const std::string& name) const override;
+
+    /// @brief Returns 'true' if all components of the array are unique
+    bool
+      is_unique() const override;
+
+    /// @brief Returns an array where all components are unique
+    const ContentPtr
+      unique() const override;
+
+    /// @brief Returns 'true' if subranges are equal
+    bool
+      is_subrange_equal(const Index64& starts, const Index64& stops) const override;
+
   protected:
     /// @brief Internal function that propagates the derivation of a contiguous
     /// version of this array from one axis to the next.
@@ -758,41 +773,117 @@ namespace awkward {
                           const SliceJagged64& slicecontent,
                           const Slice& tail) const override;
 
-  /// @brief Internal function to fill JSON with boolean values.
-  void
-    tojson_boolean(ToJson& builder, bool include_beginendlist) const;
+    const ContentPtr
+      getitem_next_jagged(const Index64& slicestarts,
+                          const Index64& slicestops,
+                          const SliceVarNewAxis& slicecontent,
+                          const Slice& tail) const override;
 
-  /// @brief Internal function to fill JSON with integer values.
-  template <typename T>
-  void
-    tojson_integer(ToJson& builder, bool include_beginendlist) const;
+    /// @brief Internal function to fill JSON with boolean values.
+    void
+      tojson_boolean(ToJson& builder, bool include_beginendlist) const;
 
-  /// @brief Internal function to fill JSON with floating-point values.
-  template <typename T>
-  void
-    tojson_real(ToJson& builder, bool include_beginendlist) const;
+    /// @brief Internal function to fill JSON with integer values.
+    template <typename T>
+    void
+      tojson_integer(ToJson& builder, bool include_beginendlist) const;
 
-  /// @brief Internal function to fill JSON with string values.
-  void
-    tojson_string(ToJson& builder, bool include_beginendlist) const;
+    /// @brief Internal function to fill JSON with floating-point values.
+    template <typename T>
+    void
+      tojson_real(ToJson& builder, bool include_beginendlist) const;
+
+    /// @brief Internal function to fill JSON with complex values.
+    template <typename T>
+    void
+      tojson_complex(ToJson& builder, bool include_beginendlist) const;
+
+    /// @brief Internal function to fill JSON with string values.
+    void
+      tojson_string(ToJson& builder, bool include_beginendlist) const;
 
   private:
-  /// @brief See #ptr.
-  std::shared_ptr<void> ptr_;
-  /// @brief See #shape.
-  std::vector<ssize_t> shape_;
-  /// @brief See #strides.
-  std::vector<ssize_t> strides_;
-  /// @brief See #byteoffset.
-  ssize_t byteoffset_;
-  /// @brief See #itemsize.
-  const ssize_t itemsize_;
-  /// @brief See #format.
-  const std::string format_;
 
-  /// @brief Mapping from (platform dependent) `std::type_index` to pybind11
-  /// format string (see #format).
-  static const std::unordered_map<std::type_index, std::string> format_map;
+    /// @brief std::sort uses intro-sort
+    ///        std::stable_sort uses mergesort
+    template<typename T>
+    const std::shared_ptr<void> index_sort(const T* data,
+                                           int64_t length,
+                                           const Index64& starts,
+                                           const Index64& parents,
+                                           int64_t outlength,
+                                           bool ascending,
+                                           bool stable) const;
+
+    template<typename T>
+    const std::shared_ptr<void> array_sort(const T* data,
+                                           int64_t length,
+                                           const Index64& starts,
+                                           const Index64& parents,
+                                           int64_t outlength,
+                                           bool ascending,
+                                           bool stable) const;
+    template<typename T>
+    const std::shared_ptr<void> array_unique(const T* data,
+                                             int64_t length,
+                                             const Index64& starts,
+                                             const Index64& parents,
+                                             int64_t& outlength) const;
+
+    template<typename T>
+    const std::shared_ptr<void> string_sort(const T* data,
+                                            int64_t length,
+                                            const Index64& offsets,
+                                            Index64& outoffsets,
+                                            bool ascending,
+                                            bool stable) const;
+
+    template<typename T>
+    const std::shared_ptr<void> string_unique(const T* data,
+                                              int64_t length,
+                                              const Index64& offsets,
+                                              Index64& outoffsets,
+                                              int64_t& outlength) const;
+
+    template<typename T>
+    bool subranges_equal(const T* data,
+                         int64_t length,
+                         const Index64& starts,
+                         const Index64& stops) const;
+
+    template<typename T>
+    const std::shared_ptr<void> as_type(const T* data,
+                                        int64_t length,
+                                        const util::dtype dtype) const;
+
+    template<typename TO, typename FROM>
+    const std::shared_ptr<void> cast_to_type(const FROM* data,
+                                             int64_t length) const;
+
+    const ContentPtr
+    sort_data(bool ascending,
+              bool stable) const;
+
+    const ContentPtr
+    unique_data() const;
+
+    /// @brief See #ptr.
+    std::shared_ptr<void> ptr_;
+    /// @brief See #ptr_li
+    const kernel::lib ptr_lib_;
+    /// @brief See #shape.
+    std::vector<ssize_t> shape_;
+    /// @brief See #strides.
+    std::vector<ssize_t> strides_;
+    /// @brief See #byteoffset.
+    ssize_t byteoffset_;
+    /// @brief See #itemsize.
+    const ssize_t itemsize_;
+    /// @brief See #format.
+    const std::string format_;
+    /// @brief See #dtype.
+    const util::dtype dtype_;
+
   };
 }
 
