@@ -1,8 +1,9 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_RECORD_H_
 #define AWKWARD_RECORD_H_
 
+#include "awkward/common.h"
 #include "awkward/array/RecordArray.h"
 
 namespace awkward {
@@ -12,7 +13,7 @@ namespace awkward {
   ///
   /// Many of the methods raise runtime errors. See
   /// {@link Content#isscalar Content::isscalar}.
-  class EXPORT_SYMBOL Record: public Content {
+  class LIBAWKWARD_EXPORT_SYMBOL Record: public Content {
   public:
     /// @brief Creates a Record from a full set of parameters.
     ///
@@ -76,11 +77,11 @@ namespace awkward {
     const FormPtr
       form(bool materialize) const override;
 
-    bool
-      has_virtual_form() const override;
+    kernel::lib
+      kernels() const override;
 
-    bool
-      has_virtual_length() const override;
+    void
+      caches(std::vector<ArrayCachePtr>& out) const override;
 
     const std::string
       tostring_part(const std::string& indent,
@@ -130,12 +131,23 @@ namespace awkward {
     const ContentPtr
       getitem_field(const std::string& key) const override;
 
+    const ContentPtr
+      getitem_field(const std::string& key,
+                    const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_fields(const std::vector<std::string>& keys) const override;
 
+    const ContentPtr
+      getitem_fields(const std::vector<std::string>& keys,
+                     const Slice& only_fields) const override;
+
     /// @exception std::runtime_error is always thrown
     const ContentPtr
-      carry(const Index64& carry) const override;
+      carry(const Index64& carry, bool allow_lazy) const override;
+
+    bool
+      dimension_optiontype() const override;
 
     int64_t
       purelist_depth() const override;
@@ -180,9 +192,12 @@ namespace awkward {
     bool
       mergeable(const ContentPtr& other, bool mergebool) const override;
 
+    bool
+      referentially_equal(const ContentPtr& other) const override;
+
     /// @exception std::runtime_error is always thrown
     const ContentPtr
-      merge(const ContentPtr& other) const override;
+      mergemany(const ContentPtrVec& others) const override;
 
     /// @exception std::runtime_error is always thrown
     const SliceItemPtr
@@ -205,11 +220,30 @@ namespace awkward {
       reduce_next(const Reducer& reducer,
                   int64_t negaxis,
                   const Index64& starts,
+                  const Index64& shifts,
                   const Index64& parents,
                   int64_t outlength,
                   bool mask,
                   bool keepdims) const override;
 
+    const ContentPtr
+      sort_next(int64_t negaxis,
+                const Index64& starts,
+                const Index64& parents,
+                int64_t outlength,
+                bool ascending,
+                bool stable,
+                bool keepdims) const override;
+
+    const ContentPtr
+      argsort_next(int64_t negaxis,
+                   const Index64& starts,
+                   const Index64& parents,
+                   int64_t outlength,
+                   bool ascending,
+                   bool stable,
+                   bool keepdims) const override;
+
     const ContentPtr
       localindex(int64_t axis, int64_t depth) const override;
 
@@ -308,6 +342,39 @@ namespace awkward {
                           const SliceJagged64& slicecontent,
                           const Slice& tail) const override;
 
+    /// @exception std::runtime_error is always thrown
+    const ContentPtr
+      getitem_next_jagged(const Index64& slicestarts,
+                          const Index64& slicestops,
+                          const SliceVarNewAxis& slicecontent,
+                          const Slice& tail) const override;
+
+    const ContentPtr
+      getitem_next(const SliceVarNewAxis& varnewaxis,
+                   const Slice& tail,
+                   const Index64& advanced) const override;
+
+    const SliceJagged64
+      varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const override;
+
+    const ContentPtr
+      copy_to(kernel::lib ptr_lib) const override;
+
+    const ContentPtr
+      numbers_to_type(const std::string& name) const override;
+
+    /// @brief Returns 'true' if all components of the array are unique
+    bool
+      is_unique() const override;
+
+    /// @brief Returns an array where all components are unique
+    const ContentPtr
+      unique() const override;
+
+    /// @brief Returns 'true' if subranges are equal
+    bool
+      is_subrange_equal(const Index64& start, const Index64& stop) const override;
+
   private:
     /// @brief See #array.
     const std::shared_ptr<const RecordArray> array_;
