@@ -1,4 +1,4 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_RECORDARRAY_H_
 #define AWKWARD_RECORDARRAY_H_
@@ -7,7 +7,7 @@
 #include <string>
 #include <unordered_map>
 
-#include "awkward/cpu-kernels/util.h"
+#include "awkward/common.h"
 #include "awkward/Identities.h"
 #include "awkward/Content.h"
 
@@ -15,12 +15,13 @@ namespace awkward {
   /// @class RecordForm
   ///
   /// @brief Form describing RecordArray (not a Record).
-  class EXPORT_SYMBOL RecordForm: public Form {
+  class LIBAWKWARD_EXPORT_SYMBOL RecordForm: public Form {
   public:
     /// @brief Creates a RecordForm. See RecordArray (not Record) for
     /// documentation.
     RecordForm(bool has_identities,
                const util::Parameters& parameters,
+               const FormKey& form_key,
                const util::RecordLookupPtr& recordlookup,
                const std::vector<FormPtr>& contents);
 
@@ -51,6 +52,9 @@ namespace awkward {
     const FormPtr
       shallow_copy() const override;
 
+    const FormPtr
+      with_form_key(const FormKey& form_key) const override;
+
     const std::string
       purelist_parameter(const std::string& key) const override;
 
@@ -60,6 +64,9 @@ namespace awkward {
     int64_t
       purelist_depth() const override;
 
+    bool
+      dimension_optiontype() const override;
+
     const std::pair<int64_t, int64_t>
       minmax_depth() const override;
 
@@ -84,7 +91,15 @@ namespace awkward {
     bool
       equal(const FormPtr& other,
             bool check_identities,
-            bool check_parameters) const override;
+            bool check_parameters,
+            bool check_form_key,
+            bool compatibility_check) const override;
+
+    const FormPtr
+      getitem_field(const std::string& key) const override;
+
+    const FormPtr
+      getitem_fields(const std::vector<std::string>& keys) const override;
 
   private:
     const util::RecordLookupPtr recordlookup_;
@@ -104,7 +119,7 @@ namespace awkward {
   /// with each field.
   ///
   /// Fields are always ordered, whether tuples or records.
-  class EXPORT_SYMBOL RecordArray:
+  class LIBAWKWARD_EXPORT_SYMBOL RecordArray:
     public Content,
     public std::enable_shared_from_this<RecordArray> {
   public:
@@ -118,10 +133,23 @@ namespace awkward {
     /// (ordered) fields.
     /// @param recordlookup A `std::shared_ptr<std::vector<std::string>>`
     /// optional list of key names.
+    /// @param length The length of the RecordArray; must be less than or
+    /// equal to the lengths of all #contents.
+    /// @param caches A list of all #VirtualArray caches contained within
+    /// this structure, so that they don't have to be searched repeatedly.
     /// If absent (`nullptr`), the data are tuples; otherwise, they are
     /// records. The number of names must match the number of #contents.
     /// @param length The length of the array, breaking ambiguities between
     /// #contents of different lengths and essential if there are zero fields.
+    RecordArray(const IdentitiesPtr& identities,
+                const util::Parameters& parameters,
+                const ContentPtrVec& contents,
+                const util::RecordLookupPtr& recordlookup,
+                int64_t length,
+                const std::vector<ArrayCachePtr>& caches);
+
+    /// @brief Creates a RecordArray in which #caches is derived from
+    /// the #contents.
     RecordArray(const IdentitiesPtr& identities,
                 const util::Parameters& parameters,
                 const ContentPtrVec& contents,
@@ -180,11 +208,11 @@ namespace awkward {
     const FormPtr
       form(bool materialize) const override;
 
-    bool
-      has_virtual_form() const override;
+    kernel::lib
+      kernels() const override;
 
-    bool
-      has_virtual_length() const override;
+    void
+      caches(std::vector<ArrayCachePtr>& out) const override;
 
     const std::string
       tostring_part(const std::string& indent,
@@ -235,16 +263,33 @@ namespace awkward {
     const ContentPtr
       getitem_field(const std::string& key) const override;
 
+    const ContentPtr
+      getitem_field(const std::string& key,
+                    const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_fields(const std::vector<std::string>& keys) const override;
 
+    const ContentPtr
+      getitem_fields(const std::vector<std::string>& keys,
+                     const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_next(const SliceItemPtr& head,
                    const Slice& tail,
                    const Index64& advanced) const override;
 
     const ContentPtr
-      carry(const Index64& carry) const override;
+      carry(const Index64& carry, bool allow_lazy) const override;
+
+    int64_t
+      purelist_depth() const override;
+
+    const std::pair<int64_t, int64_t>
+      minmax_depth() const override;
+
+    const std::pair<bool, int64_t>
+      branch_depth() const override;
 
     int64_t
       numfields() const override;
@@ -280,8 +325,11 @@ namespace awkward {
     bool
       mergeable(const ContentPtr& other, bool mergebool) const override;
 
+    bool
+      referentially_equal(const ContentPtr& other) const override;
+
     const ContentPtr
-      merge(const ContentPtr& other) const override;
+      mergemany(const ContentPtrVec& others) const override;
 
     const SliceItemPtr
       asslice() const override;
@@ -301,11 +349,30 @@ namespace awkward {
       reduce_next(const Reducer& reducer,
                   int64_t negaxis,
                   const Index64& starts,
+                  const Index64& shifts,
                   const Index64& parents,
                   int64_t outlength,
                   bool mask,
                   bool keepdims) const override;
 
+    const ContentPtr
+      sort_next(int64_t negaxis,
+                const Index64& starts,
+                const Index64& parents,
+                int64_t outlength,
+                bool ascending,
+                bool stable,
+                bool keepdims) const override;
+
+    const ContentPtr
+      argsort_next(int64_t negaxis,
+                   const Index64& starts,
+                   const Index64& parents,
+                   int64_t outlength,
+                   bool ascending,
+                   bool stable,
+                   bool keepdims) const override;
+
     const ContentPtr
       localindex(int64_t axis, int64_t depth) const override;
 
@@ -396,6 +463,38 @@ namespace awkward {
                           const SliceJagged64& slicecontent,
                           const Slice& tail) const override;
 
+    const ContentPtr
+      getitem_next_jagged(const Index64& slicestarts,
+                          const Index64& slicestops,
+                          const SliceVarNewAxis& slicecontent,
+                          const Slice& tail) const override;
+
+    const ContentPtr
+      getitem_next(const SliceVarNewAxis& varnewaxis,
+                   const Slice& tail,
+                   const Index64& advanced) const override;
+
+    const SliceJagged64
+      varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const override;
+
+    const ContentPtr
+      copy_to(kernel::lib ptr_lib) const override;
+
+    const ContentPtr
+      numbers_to_type(const std::string& name) const override;
+
+    /// @brief Returns 'true' if all components of the array are unique
+    bool
+      is_unique() const override;
+
+    /// @brief Returns an array where all components are unique
+    const ContentPtr
+      unique() const override;
+
+    /// @brief Returns 'true' if subranges are equal
+    bool
+      is_subrange_equal(const Index64& start, const Index64& stop) const override;
+
   protected:
     template <typename S>
     const ContentPtr
@@ -411,6 +510,8 @@ namespace awkward {
     const util::RecordLookupPtr recordlookup_;
     /// @brief See #length.
     int64_t length_;
+    /// @brief See #caches.
+    const std::vector<ArrayCachePtr> caches_;
   };
 }
 
