@@ -1,4 +1,4 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
 #ifndef AWKWARD_REGULARARRAY_H_
 #define AWKWARD_REGULARARRAY_H_
@@ -7,7 +7,7 @@
 #include <memory>
 #include <vector>
 
-#include "awkward/cpu-kernels/util.h"
+#include "awkward/common.h"
 #include "awkward/Slice.h"
 #include "awkward/Content.h"
 
@@ -15,11 +15,12 @@ namespace awkward {
   /// @class RegularForm
   ///
   /// @brief Form describing RegularArray.
-  class EXPORT_SYMBOL RegularForm: public Form {
+  class LIBAWKWARD_EXPORT_SYMBOL RegularForm: public Form {
   public:
     /// @brief Creates a RegularForm. See RegularArray for documentation.
     RegularForm(bool has_identities,
                 const util::Parameters& parameters,
+                const FormKey& form_key,
                 const FormPtr& content,
                 int64_t size);
 
@@ -38,6 +39,9 @@ namespace awkward {
     const FormPtr
       shallow_copy() const override;
 
+    const FormPtr
+      with_form_key(const FormKey& form_key) const override;
+
     const std::string
       purelist_parameter(const std::string& key) const override;
 
@@ -47,6 +51,9 @@ namespace awkward {
     int64_t
       purelist_depth() const override;
 
+    bool
+      dimension_optiontype() const override;
+
     const std::pair<int64_t, int64_t>
       minmax_depth() const override;
 
@@ -71,7 +78,15 @@ namespace awkward {
     bool
       equal(const FormPtr& other,
             bool check_identities,
-            bool check_parameters) const override;
+            bool check_parameters,
+            bool check_form_key,
+            bool compatibility_check) const override;
+
+    const FormPtr
+      getitem_field(const std::string& key) const override;
+
+    const FormPtr
+      getitem_fields(const std::vector<std::string>& keys) const override;
 
   private:
     const FormPtr content_;
@@ -87,7 +102,7 @@ namespace awkward {
   /// it must also start at zero.
   ///
   /// See #RegularArray for the meaning of each parameter.
-  class EXPORT_SYMBOL RegularArray: public Content {
+  class LIBAWKWARD_EXPORT_SYMBOL RegularArray: public Content {
   public:
     /// @brief Creates a RegularArray from a full set of parameters.
     ///
@@ -100,6 +115,15 @@ namespace awkward {
     /// Values in `content[i]` where `i >= length * size` are "unreachable,"
     /// and don't exist in the high level view.
     /// @param size Length of the equally sized nested lists.
+    /// @param zeros_length If `size == 0`, this is the length of the array; otherwise,
+    /// the length of the array is derived as the floor of length / size.
+    RegularArray(const IdentitiesPtr& identities,
+                 const util::Parameters& parameters,
+                 const ContentPtr& content,
+                 int64_t size,
+                 int64_t zeros_length);
+
+    /// @brief Creates a RegularArray in which `zeros_length` is zero.
     RegularArray(const IdentitiesPtr& identities,
                  const util::Parameters& parameters,
                  const ContentPtr& content,
@@ -166,11 +190,11 @@ namespace awkward {
     const FormPtr
       form(bool materialize) const override;
 
-    bool
-      has_virtual_form() const override;
+    kernel::lib
+      kernels() const override;
 
-    bool
-      has_virtual_length() const override;
+    void
+      caches(std::vector<ArrayCachePtr>& out) const override;
 
     const std::string
       tostring_part(const std::string& indent,
@@ -185,7 +209,7 @@ namespace awkward {
 
     /// @copydoc Content::length()
     ///
-    /// Equal to `floor(len(content) / size)`.
+    /// Equal to `floor(len(content) / size)` or `0` if `size == 0`.
     int64_t
       length() const override;
 
@@ -218,11 +242,28 @@ namespace awkward {
     const ContentPtr
       getitem_field(const std::string& key) const override;
 
+    const ContentPtr
+      getitem_field(const std::string& key,
+                    const Slice& only_fields) const override;
+
     const ContentPtr
       getitem_fields(const std::vector<std::string>& keys) const override;
 
     const ContentPtr
-      carry(const Index64& carry) const override;
+      getitem_fields(const std::vector<std::string>& keys,
+                     const Slice& only_fields) const override;
+
+    const ContentPtr
+      carry(const Index64& carry, bool allow_lazy) const override;
+
+    int64_t
+      purelist_depth() const override;
+
+    const std::pair<int64_t, int64_t>
+      minmax_depth() const override;
+
+    const std::pair<bool, int64_t>
+      branch_depth() const override;
 
     int64_t
       numfields() const override;
@@ -258,8 +299,11 @@ namespace awkward {
     bool
       mergeable(const ContentPtr& other, bool mergebool) const override;
 
+    bool
+      referentially_equal(const ContentPtr& other) const override;
+
     const ContentPtr
-      merge(const ContentPtr& other) const override;
+      mergemany(const ContentPtrVec& others) const override;
 
     const SliceItemPtr
       asslice() const override;
@@ -279,11 +323,30 @@ namespace awkward {
       reduce_next(const Reducer& reducer,
                   int64_t negaxis,
                   const Index64& starts,
+                  const Index64& shifts,
                   const Index64& parents,
                   int64_t outlength,
                   bool mask,
                   bool keepdims) const override;
 
+    const ContentPtr
+      sort_next(int64_t negaxis,
+                const Index64& starts,
+                const Index64& parents,
+                int64_t outlength,
+                bool ascending,
+                bool stable,
+                bool keepdims) const override;
+
+    const ContentPtr
+      argsort_next(int64_t negaxis,
+                   const Index64& starts,
+                   const Index64& parents,
+                   int64_t outlength,
+                   bool ascending,
+                   bool stable,
+                   bool keepdims) const override;
+
     const ContentPtr
       localindex(int64_t axis, int64_t depth) const override;
 
@@ -333,9 +396,42 @@ namespace awkward {
                           const SliceJagged64& slicecontent,
                           const Slice& tail) const override;
 
+    const ContentPtr
+      getitem_next_jagged(const Index64& slicestarts,
+                          const Index64& slicestops,
+                          const SliceVarNewAxis& slicecontent,
+                          const Slice& tail) const override;
+
+    const ContentPtr
+      getitem_next(const SliceVarNewAxis& varnewaxis,
+                   const Slice& tail,
+                   const Index64& advanced) const override;
+
+    const SliceJagged64
+      varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const override;
+
+    const ContentPtr
+      copy_to(kernel::lib ptr_lib) const override;
+
+    const ContentPtr
+      numbers_to_type(const std::string& name) const override;
+
+    /// @brief Returns 'true' if all components of the array are unique
+    bool
+      is_unique() const override;
+
+    /// @brief Returns an array where all components are unique
+    const ContentPtr
+      unique() const override;
+
+    /// @brief Returns 'true' if subranges are equal
+    bool
+      is_subrange_equal(const Index64& start, const Index64& stop) const override;
+
   private:
     const ContentPtr content_;
     int64_t size_;
+    int64_t length_;
   };
 }
 
