@@ -1,10 +1,12 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
+
+#define FILENAME(line) FILENAME_FOR_EXCEPTIONS("src/libawkward/Slice.cpp", line)
 
 #include <algorithm>
 #include <sstream>
 #include <type_traits>
 
-#include "awkward/cpu-kernels/getitem.h"
+#include "awkward/kernels.h"
 #include "awkward/util.h"
 
 #define AWKWARD_SLICE_NO_EXTERN_TEMPLATE
@@ -15,6 +17,11 @@ namespace awkward {
 
   SliceItem::~SliceItem() = default;
 
+  const SliceItemPtr
+  SliceItem::carry(const Index64& carry) const {
+    return shallow_copy();
+  }
+
   ////////// SliceAt
 
   SliceAt::SliceAt(int64_t at)
@@ -40,6 +47,16 @@ namespace awkward {
     return false;
   }
 
+  bool
+  SliceAt::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceAt* raw = dynamic_cast<SliceAt*>(other.get())) {
+      return at_ == raw->at();
+    }
+    else {
+      return false;
+    }
+  }
+
   ////////// SliceRange
 
   SliceRange::SliceRange(int64_t start, int64_t stop, int64_t step)
@@ -47,7 +64,8 @@ namespace awkward {
       , stop_(stop)
       , step_(step == Slice::none() ? 1 : step) {
     if (step_ == 0) {
-      throw std::runtime_error("step must not be zero");
+      throw std::runtime_error(
+        std::string("step must not be zero") + FILENAME(__LINE__));
     }
   }
 
@@ -102,6 +120,18 @@ namespace awkward {
     return true;
   }
 
+  bool
+  SliceRange::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceRange* raw = dynamic_cast<SliceRange*>(other.get())) {
+      return start_ == raw->start()  &&
+             stop_ == raw->stop()  &&
+             step_ == raw->step();
+    }
+    else {
+      return false;
+    }
+  }
+
   ////////// SliceEllipsis
 
   SliceEllipsis::SliceEllipsis() { }
@@ -121,6 +151,16 @@ namespace awkward {
     return true;
   }
 
+  bool
+  SliceEllipsis::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceEllipsis* raw = dynamic_cast<SliceEllipsis*>(other.get())) {
+      return true;
+    }
+    else {
+      return false;
+    }
+  }
+
   ////////// SliceNewAxis
 
   SliceNewAxis::SliceNewAxis() { }
@@ -140,6 +180,16 @@ namespace awkward {
     return false;
   }
 
+  bool
+  SliceNewAxis::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceNewAxis* raw = dynamic_cast<SliceNewAxis*>(other.get())) {
+      return true;
+    }
+    else {
+      return false;
+    }
+  }
+
   ////////// SliceArrayOf<T>
 
   template <typename T>
@@ -152,11 +202,13 @@ namespace awkward {
       , strides_(strides)
       , frombool_(frombool) {
     if (shape_.empty()) {
-      throw std::runtime_error("shape must not be zero-dimensional");
+      throw std::runtime_error(
+        std::string("shape must not be zero-dimensional") + FILENAME(__LINE__));
     }
     if (shape_.size() != strides_.size()) {
       throw std::runtime_error(
-        "shape must have the same number of dimensions as strides");
+        std::string("shape must have the same number of dimensions as strides")
+        + FILENAME(__LINE__));
     }
   }
 
@@ -205,6 +257,23 @@ namespace awkward {
                                              frombool_);
   }
 
+  template <typename T>
+  const SliceItemPtr
+  SliceArrayOf<T>::carry(const Index64& carry) const {
+    IndexOf<T> nextindex(carry.length());
+    struct Error err = kernel::Index_carry_64<T>(
+      kernel::lib::cpu,   // DERIVE
+      nextindex.data(),
+      index_.data(),
+      carry.data(),
+      index_.length(),
+      carry.length());
+    util::handle_error(err, "SliceArrayOf<T>", nullptr);
+    std::vector<int64_t> nextshape({ carry.length() });
+    nextshape.insert(nextshape.end(), shape_.begin() + 1, shape_.end());
+    return std::make_shared<SliceArrayOf<T>>(nextindex, nextshape, strides_, frombool_);
+  }
+
   template <typename T>
   const std::string
   SliceArrayOf<T>::tostring() const {
@@ -217,7 +286,7 @@ namespace awkward {
     std::stringstream out;
     out << "[";
     if (shape_.size() == 1) {
-      if (shape_[0] < 6) {
+      if (shape_[0] <= 20) {
         for (int64_t i = 0;  i < shape_[0];  i++) {
           if (i != 0) {
             out << ", ";
@@ -226,15 +295,15 @@ namespace awkward {
         }
       }
       else {
-        for (int64_t i = 0;  i < 3;  i++) {
+        for (int64_t i = 0;  i < 10;  i++) {
           if (i != 0) {
             out << ", ";
           }
           out << (T)index_.getitem_at_nowrap(i*strides_[0]);
         }
         out << ", ..., ";
-        for (int64_t i = shape_[0] - 3;  i < shape_[0];  i++) {
-          if (i != shape_[0] - 3) {
+        for (int64_t i = shape_[0] - 10;  i < shape_[0];  i++) {
+          if (i != shape_[0] - 10) {
             out << ", ";
           }
           out << (T)index_.getitem_at_nowrap(i*strides_[0]);
@@ -244,37 +313,40 @@ namespace awkward {
     else {
       std::vector<int64_t> shape(shape_.begin() + 1, shape_.end());
       std::vector<int64_t> strides(strides_.begin() + 1, strides_.end());
-      if (shape_[0] < 6) {
+      if (shape_[0] <= 20) {
         for (int64_t i = 0;  i < shape_[0];  i++) {
           if (i != 0) {
             out << ", ";
           }
           IndexOf<T> index(index_.ptr(),
                            index_.offset() + i*strides_[0],
-                           shape_[1]);
+                           shape_[1],
+                           index_.ptr_lib());
           SliceArrayOf<T> subarray(index, shape, strides, frombool_);
           out << subarray.tostring_part();
         }
       }
       else {
-        for (int64_t i = 0;  i < 3;  i++) {
+        for (int64_t i = 0;  i < 10;  i++) {
           if (i != 0) {
             out << ", ";
           }
           IndexOf<T> index(index_.ptr(),
                            index_.offset() + i*strides_[0],
-                           shape_[1]);
+                           shape_[1],
+                           index_.ptr_lib());
           SliceArrayOf<T> subarray(index, shape, strides, frombool_);
           out << subarray.tostring_part();
         }
         out << ", ..., ";
-        for (int64_t i = shape_[0] - 3;  i < shape_[0];  i++) {
-          if (i != shape_[0] - 3) {
+        for (int64_t i = shape_[0] - 10;  i < shape_[0];  i++) {
+          if (i != shape_[0] - 10) {
             out << ", ";
           }
           IndexOf<T> index(index_.ptr(),
                            index_.offset() + i*strides_[0],
-                           shape_[1]);
+                           shape_[1],
+                           index_.ptr_lib());
           SliceArrayOf<T> subarray(index, shape, strides, frombool_);
           out << subarray.tostring_part();
         }
@@ -290,6 +362,19 @@ namespace awkward {
     return advanced.length() == 0;
   }
 
+  template <typename T>
+  bool
+  SliceArrayOf<T>::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceArrayOf<T>* raw = dynamic_cast<SliceArrayOf<T>*>(other.get())) {
+      return shape_ == raw->shape()  &&
+             strides_ == raw->strides()  &&
+             index_.referentially_equal(raw->index());
+    }
+    else {
+      return false;
+    }
+  }
+
   template <typename T>
   const IndexOf<T>
   SliceArrayOf<T>::ravel() const {
@@ -300,20 +385,25 @@ namespace awkward {
 
     IndexOf<T> index(length);
     if (std::is_same<T, int64_t>::value) {
-      awkward_slicearray_ravel_64(index.ptr().get(),
-                                  index_.ptr().get(),
-                                  ndim(),
-                                  shape_.data(),
-                                  strides_.data());
+      struct Error err = kernel::slicearray_ravel_64(
+        kernel::lib::cpu,   // DERIVE
+        index.ptr().get(),
+        index_.ptr().get(),
+        ndim(),
+        shape_.data(),
+        strides_.data());
+      util::handle_error(err);
     }
     else {
-      throw std::runtime_error("unrecognized SliceArrayOf<T> type");
+      throw std::runtime_error(
+        std::string("unrecognized SliceArrayOf<T> type")
+        + FILENAME(__LINE__));
     }
 
     return index;
   }
 
-  template class SliceArrayOf<int64_t>;
+  template class EXPORT_TEMPLATE_INST SliceArrayOf<int64_t>;
 
   ////////// SliceField
 
@@ -332,7 +422,7 @@ namespace awkward {
 
   const std::string
   SliceField::tostring() const {
-    return util::quote(key_, true);
+    return util::quote(key_);
   }
 
   bool
@@ -340,6 +430,16 @@ namespace awkward {
     return false;
   }
 
+  bool
+  SliceField::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceField* raw = dynamic_cast<SliceField*>(other.get())) {
+      return key_ == raw->key();
+    }
+    else {
+      return false;
+    }
+  }
+
   ////////// SliceFields
 
   SliceFields::SliceFields(const std::vector<std::string>& keys)
@@ -363,7 +463,7 @@ namespace awkward {
       if (i != 0) {
         out << ", ";
       }
-      out << util::quote(keys_[i], true);
+      out << util::quote(keys_[i]);
     }
     out << "]";
     return out.str();
@@ -374,6 +474,16 @@ namespace awkward {
     return false;
   }
 
+  bool
+  SliceFields::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceFields* raw = dynamic_cast<SliceFields*>(other.get())) {
+      return keys_ == raw->keys();
+    }
+    else {
+      return false;
+    }
+  }
+
   ////////// SliceMissingOf<T>
 
   template <typename T>
@@ -416,6 +526,49 @@ namespace awkward {
                                                content_);
   }
 
+  template <typename T>
+  const SliceItemPtr
+  SliceMissingOf<T>::carry(const Index64& carry) const {
+    IndexOf<T> nextindex(carry.length());
+    struct Error err1 = kernel::Index_carry_64<T>(
+      kernel::lib::cpu,   // DERIVE
+      nextindex.data(),
+      index_.data(),
+      carry.data(),
+      index_.length(),
+      carry.length());
+    util::handle_error(err1, "SliceMissingOf<T>", nullptr);
+
+    int64_t numnull;
+    struct Error err2 = kernel::IndexedArray_numnull<T>(
+      kernel::lib::cpu,   // DERIVE
+      &numnull,
+      nextindex.data(),
+      nextindex.length());
+    util::handle_error(err2, "SliceMissingOf<T>", nullptr);
+
+    Index64 nextcarry(nextindex.length() - numnull);
+    struct Error err3 = kernel::IndexedArray_flatten_nextcarry_64<T>(
+      kernel::lib::cpu,   // DERIVE
+      nextcarry.data(),
+      nextindex.data(),
+      nextindex.length(),
+      kMaxInt64);
+    util::handle_error(err3, "SliceMissingOf<T>", nullptr);
+
+    SliceItemPtr nextcontent = content_.get()->carry(nextcarry);
+    IndexOf<T> outindex(nextindex.length());
+
+    struct Error err = kernel::carry_SliceMissing_outindex<T>(
+      kernel::lib::cpu,   // DERIVE
+      outindex.data(),
+      nextindex.data(),
+      nextindex.length());
+    util::handle_error(err, "SliceJaggedOf<T>", nullptr);
+
+    return std::make_shared<SliceMissingOf<T>>(outindex, originalmask_, nextcontent);
+  }
+
   template <typename T>
   const std::string
   SliceMissingOf<T>::tostring() const {
@@ -428,7 +581,7 @@ namespace awkward {
   SliceMissingOf<T>::tostring_part() const {
     std::stringstream out;
     out << "[";
-    if (index_.length() < 6) {
+    if (index_.length() <= 20) {
       for (int64_t i = 0;  i < index_.length();  i++) {
         if (i != 0) {
           out << ", ";
@@ -437,15 +590,15 @@ namespace awkward {
       }
     }
     else {
-      for (int64_t i = 0;  i < 3;  i++) {
+      for (int64_t i = 0;  i < 10;  i++) {
         if (i != 0) {
           out << ", ";
         }
         out << (T)index_.getitem_at_nowrap(i);
       }
       out << ", ..., ";
-      for (int64_t i = index_.length() - 3;  i < index_.length();  i++) {
-        if (i != index_.length() - 3) {
+      for (int64_t i = index_.length() - 10;  i < index_.length();  i++) {
+        if (i != index_.length() - 10) {
           out << ", ";
         }
         out << (T)index_.getitem_at_nowrap(i);
@@ -461,7 +614,19 @@ namespace awkward {
     return true;
   }
 
-  template class SliceMissingOf<int64_t>;
+  template <typename T>
+  bool
+  SliceMissingOf<T>::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceMissingOf<T>* raw = dynamic_cast<SliceMissingOf<T>*>(other.get())) {
+      return index_.referentially_equal(raw->index())  &&
+             content_.get()->referentially_equal(raw->content());
+    }
+    else {
+      return false;
+    }
+  }
+
+  template class EXPORT_TEMPLATE_INST SliceMissingOf<int64_t>;
 
   ////////// SliceJaggedOf<T>
 
@@ -495,6 +660,33 @@ namespace awkward {
     return std::make_shared<SliceJaggedOf<T>>(offsets_, content_);
   }
 
+  template <typename T>
+  const SliceItemPtr
+  SliceJaggedOf<T>::carry(const Index64& carry) const {
+    IndexOf<T> nextoffsets(carry.length() + 1);
+    int64_t nextcontentlen = 0;
+
+    struct Error err1 = kernel::carry_SliceJagged_offsets<T>(
+      kernel::lib::cpu,   // DERIVE
+      nextoffsets.data(),
+      offsets_.data(),
+      carry.data(),
+      carry.length());
+    util::handle_error(err1, "SliceJaggedOf<T>", nullptr);
+
+    Index64 nextcarry(nextoffsets.getitem_at_nowrap(carry.length()));
+    struct Error err2 = kernel::carry_SliceJagged_nextcarry<T>(
+      kernel::lib::cpu,   // DERIVE
+      nextcarry.data(),
+      offsets_.data(),
+      carry.data(),
+      carry.length());
+    util::handle_error(err2, "SliceJaggedOf<T>", nullptr);
+
+    SliceItemPtr nextcontent = content_.get()->carry(nextcarry);
+    return std::make_shared<SliceJaggedOf<T>>(nextoffsets, nextcontent);
+  }
+
   template <typename T>
   const std::string
   SliceJaggedOf<T>::tostring() const {
@@ -507,7 +699,7 @@ namespace awkward {
   SliceJaggedOf<T>::tostring_part() const {
     std::stringstream out;
     out << "[";
-    if (offsets_.length() < 6) {
+    if (offsets_.length() <= 20) {
       for (int64_t i = 0;  i < offsets_.length();  i++) {
         if (i != 0) {
           out << ", ";
@@ -516,15 +708,15 @@ namespace awkward {
       }
     }
     else {
-      for (int64_t i = 0;  i < 3;  i++) {
+      for (int64_t i = 0;  i < 10;  i++) {
         if (i != 0) {
           out << ", ";
         }
         out << (T)offsets_.getitem_at_nowrap(i);
       }
       out << ", ..., ";
-      for (int64_t i = offsets_.length() - 3;  i < offsets_.length();  i++) {
-        if (i != offsets_.length() - 3) {
+      for (int64_t i = offsets_.length() - 10;  i < offsets_.length();  i++) {
+        if (i != offsets_.length() - 10) {
           out << ", ";
         }
         out << (T)offsets_.getitem_at_nowrap(i);
@@ -540,7 +732,59 @@ namespace awkward {
     return true;
   }
 
-  template class SliceJaggedOf<int64_t>;
+  template <typename T>
+  bool
+  SliceJaggedOf<T>::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceJaggedOf<T>* raw = dynamic_cast<SliceJaggedOf<T>*>(other.get())) {
+      return offsets_.referentially_equal(raw->offsets())  &&
+             content_.get()->referentially_equal(raw->content());
+    }
+    else {
+      return false;
+    }
+  }
+
+  template class EXPORT_TEMPLATE_INST SliceJaggedOf<int64_t>;
+
+  ////////// SliceVarNewAxis
+
+  SliceVarNewAxis::SliceVarNewAxis(const SliceItemPtr& content)
+      : content_(content) { }
+
+  const SliceItemPtr
+  SliceVarNewAxis::content() const {
+    return content_;
+  }
+
+  const SliceItemPtr
+  SliceVarNewAxis::shallow_copy() const {
+    return std::make_shared<SliceVarNewAxis>(content_);
+  }
+
+  const SliceItemPtr
+  SliceVarNewAxis::carry(const Index64& carry) const {
+    return std::make_shared<SliceVarNewAxis>(content_.get()->carry(carry));
+  }
+
+  const std::string
+  SliceVarNewAxis::tostring() const {
+    return std::string("newaxis(") + content_.get()->tostring() + std::string(")");
+  }
+
+  bool
+  SliceVarNewAxis::preserves_type(const Index64& advanced) const {
+    return true;
+  }
+
+  bool
+  SliceVarNewAxis::referentially_equal(const SliceItemPtr& other) const {
+    if (SliceVarNewAxis* raw = dynamic_cast<SliceVarNewAxis*>(other.get())) {
+      return content_.get()->referentially_equal(raw->content());
+    }
+    else {
+      return false;
+    }
+  }
 
   ////////// Slice
 
@@ -611,6 +855,30 @@ namespace awkward {
     return Slice(items, true);
   }
 
+  const Slice
+  Slice::only_fields() const {
+    std::vector<SliceItemPtr> items;
+    for (auto item : items_) {
+      if (dynamic_cast<SliceField*>(item.get())  ||
+          dynamic_cast<SliceFields*>(item.get())) {
+        items.push_back(item);
+      }
+    }
+    return Slice(items, true);
+  }
+
+  const Slice
+  Slice::not_fields() const {
+    std::vector<SliceItemPtr> items;
+    for (auto item : items_) {
+      if (!dynamic_cast<SliceField*>(item.get())  &&
+          !dynamic_cast<SliceFields*>(item.get())) {
+        items.push_back(item);
+      }
+    }
+    return Slice(items, true);
+  }
+
   const std::string
   Slice::tostring() const {
     std::stringstream out;
@@ -635,7 +903,9 @@ namespace awkward {
   void
   Slice::append(const SliceItemPtr& item) {
     if (sealed_) {
-      throw std::runtime_error("Slice::append when sealed_ == true");
+      throw std::runtime_error(
+        std::string("Slice::append when sealed_ == true")
+        + FILENAME(__LINE__));
     }
     items_.push_back(item);
   }
@@ -688,7 +958,9 @@ namespace awkward {
   void
   Slice::become_sealed() {
     if (sealed_) {
-      throw std::runtime_error("Slice::become_sealed when sealed_ == true");
+      throw std::runtime_error(
+        std::string("Slice::become_sealed when sealed_ == true")
+        + FILENAME(__LINE__));
     }
 
     std::vector<int64_t> shape;
@@ -697,8 +969,10 @@ namespace awkward {
         if (shape.empty()) {
           shape = array->shape();
         }
-        else if (shape.size() != array->ndim()) {
-          throw std::invalid_argument("cannot broadcast arrays in slice");
+        else if (shape.size() != (size_t)array->ndim()) {
+          throw std::invalid_argument(
+            std::string("cannot broadcast arrays in slice")
+            + FILENAME(__LINE__));
         }
         else {
           std::vector<int64_t> arrayshape = array->shape();
@@ -741,7 +1015,9 @@ namespace awkward {
               strides.push_back(0);
             }
             else {
-              throw std::invalid_argument("cannot broadcast arrays in slice");
+              throw std::invalid_argument(
+                std::string("cannot broadcast arrays in slice")
+                + FILENAME(__LINE__));
             }
           }
           items_[i] = std::make_shared<SliceArray64>(array->index(),
@@ -784,17 +1060,19 @@ namespace awkward {
 
       if (std::count(types.begin(), types.end(), '.') > 1) {
         throw std::invalid_argument(
-          "a slice can have no more than one ellipsis ('...')");
+          std::string("a slice can have no more than one ellipsis ('...')")
+          + FILENAME(__LINE__));
       }
 
-      size_t numadvanced = std::count(types.begin(), types.end(), 'A');
+      size_t numadvanced = (size_t)std::count(types.begin(), types.end(), 'A');
       if (numadvanced != 0) {
-        types = types.substr(0, types.find_last_of("A") + 1)
-                     .substr(types.find_first_of("A"));
+        types = types.substr(0, types.find_last_of('A') + 1)
+                     .substr(types.find_first_of('A'));
         if (numadvanced != types.size()) {
           throw std::invalid_argument(
-            "advanced indexes separated by basic indexes is not permitted "
-            "(simple integers are advanced when any arrays are present)");
+            std::string("advanced indexes separated by basic indexes is not permitted "
+                        "(simple integers are advanced when any arrays are present)")
+            + FILENAME(__LINE__));
         }
       }
     }
@@ -805,7 +1083,9 @@ namespace awkward {
   bool
   Slice::isadvanced() const {
     if (!sealed_) {
-      throw std::runtime_error("Slice::isadvanced when sealed_ == false");
+      throw std::runtime_error(
+        std::string("Slice::isadvanced when sealed_ == false")
+        + FILENAME(__LINE__));
     }
     for (size_t i = 0;  i < items_.size();  i++) {
       if (dynamic_cast<SliceArray64*>(items_[i].get()) != nullptr) {
@@ -814,4 +1094,21 @@ namespace awkward {
     }
     return false;
   }
+
+  bool
+  Slice::referentially_equal(const Slice& other) const {
+    std::vector<SliceItemPtr> others = other.items();
+
+    if (items_.size() != others.size()) {
+      return false;
+    }
+
+    for (size_t i = 0;  i < items_.size();  i++) {
+      if (!items_[i].get()->referentially_equal(others[i])) {
+        return false;
+      }
+    }
+    return true;
+  }
+
 }
