@@ -1,15 +1,18 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
+
+#define FILENAME(line) FILENAME_FOR_EXCEPTIONS("src/libawkward/array/BitMaskedArray.cpp", line)
+#define FILENAME_C(line) FILENAME_FOR_EXCEPTIONS_C("src/libawkward/array/BitMaskedArray.cpp", line)
 
 #include <sstream>
 #include <type_traits>
 
-#include "awkward/cpu-kernels/identities.h"
-#include "awkward/cpu-kernels/getitem.h"
-#include "awkward/cpu-kernels/operations.h"
+#include "awkward/kernels.h"
+#include "awkward/kernel-utils.h"
 #include "awkward/type/OptionType.h"
-#include "awkward/type/ArrayType.h"
 #include "awkward/type/UnknownType.h"
+#include "awkward/Reducer.h"
 #include "awkward/Slice.h"
+#include "awkward/io/json.h"
 #include "awkward/array/None.h"
 #include "awkward/array/EmptyArray.h"
 #include "awkward/array/UnionArray.h"
@@ -26,11 +29,12 @@ namespace awkward {
 
   BitMaskedForm::BitMaskedForm(bool has_identities,
                                const util::Parameters& parameters,
+                               const FormKey& form_key,
                                Index::Form mask,
                                const FormPtr& content,
                                bool valid_when,
                                bool lsb_order)
-      : Form(has_identities, parameters)
+      : Form(has_identities, parameters, form_key)
       , mask_(mask)
       , content_(content)
       , valid_when_(valid_when)
@@ -79,6 +83,7 @@ namespace awkward {
     builder.boolean(lsb_order_);
     identities_tojson(builder, verbose);
     parameters_tojson(builder, verbose);
+    form_key_tojson(builder, verbose);
     builder.endrecord();
   }
 
@@ -86,6 +91,18 @@ namespace awkward {
   BitMaskedForm::shallow_copy() const {
     return std::make_shared<BitMaskedForm>(has_identities_,
                                            parameters_,
+                                           form_key_,
+                                           mask_,
+                                           content_,
+                                           valid_when_,
+                                           lsb_order_);
+  }
+
+  const FormPtr
+  BitMaskedForm::with_form_key(const FormKey& form_key) const {
+    return std::make_shared<BitMaskedForm>(has_identities_,
+                                           parameters_,
+                                           form_key,
                                            mask_,
                                            content_,
                                            valid_when_,
@@ -113,6 +130,11 @@ namespace awkward {
     return content_.get()->purelist_depth();
   }
 
+  bool
+  BitMaskedForm::dimension_optiontype() const {
+    return true;
+  }
+
   const std::pair<int64_t, int64_t>
   BitMaskedForm::minmax_depth() const {
     return content_.get()->minmax_depth();
@@ -151,20 +173,40 @@ namespace awkward {
   bool
   BitMaskedForm::equal(const FormPtr& other,
                        bool check_identities,
-                       bool check_parameters) const {
+                       bool check_parameters,
+                       bool check_form_key,
+                       bool compatibility_check) const {
+    if (compatibility_check) {
+      if (VirtualForm* raw = dynamic_cast<VirtualForm*>(other.get())) {
+        if (raw->form().get() != nullptr) {
+          return equal(raw->form(),
+                       check_identities,
+                       check_parameters,
+                       check_form_key,
+                       compatibility_check);
+        }
+      }
+    }
+
     if (check_identities  &&
         has_identities_ != other.get()->has_identities()) {
       return false;
     }
     if (check_parameters  &&
-        !util::parameters_equal(parameters_, other.get()->parameters())) {
+        !util::parameters_equal(parameters_, other.get()->parameters(), false)) {
+      return false;
+    }
+    if (check_form_key  &&
+        !form_key_equals(other.get()->form_key())) {
       return false;
     }
     if (BitMaskedForm* t = dynamic_cast<BitMaskedForm*>(other.get())) {
       return (mask_ == t->mask()  &&
               content_.get()->equal(t->content(),
                                     check_identities,
-                                    check_parameters)  &&
+                                    check_parameters,
+                                    check_form_key,
+                                    compatibility_check)  &&
               valid_when_ == t->valid_when()  &&
               lsb_order_ == t->lsb_order());
     }
@@ -173,6 +215,60 @@ namespace awkward {
     }
   }
 
+  const FormPtr
+  BitMaskedForm::getitem_range() const {
+    return std::make_shared<ByteMaskedForm>(
+      has_identities_,
+      parameters_,
+      FormKey(nullptr),
+      Index::Form::i8,
+      content_.get()->getitem_range(),
+      valid_when_);
+  }
+
+  const FormPtr
+  BitMaskedForm::getitem_field(const std::string& key) const {
+    BitMaskedForm step1(has_identities_,
+                        util::Parameters(),
+                        FormKey(nullptr),
+                        mask_,
+                        content_.get()->getitem_field(key),
+                        valid_when_,
+                        lsb_order_);
+    return step1.simplify_optiontype();
+  }
+
+  const FormPtr
+  BitMaskedForm::getitem_fields(const std::vector<std::string>& keys) const {
+    BitMaskedForm step1(has_identities_,
+                        util::Parameters(),
+                        FormKey(nullptr),
+                        mask_,
+                        content_.get()->getitem_fields(keys),
+                        valid_when_,
+                        lsb_order_);
+    return step1.simplify_optiontype();
+  }
+
+  const FormPtr
+  BitMaskedForm::simplify_optiontype() const {
+    if (dynamic_cast<IndexedForm*>(content_.get())         ||
+        dynamic_cast<IndexedOptionForm*>(content_.get())   ||
+        dynamic_cast<ByteMaskedForm*>(content_.get())      ||
+        dynamic_cast<BitMaskedForm*>(content_.get())       ||
+        dynamic_cast<UnmaskedForm*>(content_.get())) {
+      IndexedOptionForm step1(has_identities_,
+                              parameters_,
+                              form_key_,
+                              Index::Form::i64,
+                              content_);
+      return step1.simplify_optiontype();
+    }
+    else {
+      return shallow_copy();
+    }
+  }
+
   ////////// BitMaskedArray
 
   BitMaskedArray::BitMaskedArray(const IdentitiesPtr& identities,
@@ -191,11 +287,13 @@ namespace awkward {
     int64_t bitlength = ((length / 8) + ((length % 8) != 0));
     if (mask.length() < bitlength) {
       throw std::invalid_argument(
-        "BitMaskedArray mask must not be shorter than its ceil(length / 8.0)");
+        std::string("BitMaskedArray mask must not be shorter than its ceil(length / 8.0)")
+        + FILENAME(__LINE__));
     }
     if (content.get()->length() < length) {
       throw std::invalid_argument(
-        "BitMaskedArray content must not be shorter than its length");
+        std::string("BitMaskedArray content must not be shorter than its length")
+        + FILENAME(__LINE__));
     }
   }
 
@@ -232,12 +330,12 @@ namespace awkward {
   const Index8
   BitMaskedArray::bytemask() const {
     Index8 bytemask(mask_.length() * 8);
-    struct Error err = awkward_bitmaskedarray_to_bytemaskedarray(
-      bytemask.ptr().get(),
-      mask_.ptr().get(),
-      mask_.offset(),
+    struct Error err = kernel::BitMaskedArray_to_ByteMaskedArray(
+      kernel::lib::cpu,   // DERIVE
+      bytemask.data(),
+      mask_.data(),
       mask_.length(),
-      false,
+      valid_when_,
       lsb_order_);
     util::handle_error(err, classname(), identities_.get());
     return bytemask.getitem_range_nowrap(0, length_);
@@ -266,12 +364,12 @@ namespace awkward {
   const std::shared_ptr<ByteMaskedArray>
   BitMaskedArray::toByteMaskedArray() const {
     Index8 bytemask(mask_.length() * 8);
-    struct Error err = awkward_bitmaskedarray_to_bytemaskedarray(
-      bytemask.ptr().get(),
-      mask_.ptr().get(),
-      mask_.offset(),
+    struct Error err = kernel::BitMaskedArray_to_ByteMaskedArray(
+      kernel::lib::cpu,   // DERIVE
+      bytemask.data(),
+      mask_.data(),
       mask_.length(),
-      valid_when_,
+      false,
       lsb_order_);
     util::handle_error(err, classname(), identities_.get());
     return std::make_shared<ByteMaskedArray>(
@@ -285,10 +383,10 @@ namespace awkward {
   const std::shared_ptr<IndexedOptionArray64>
   BitMaskedArray::toIndexedOptionArray64() const {
     Index64 index(mask_.length() * 8);
-    struct Error err = awkward_bitmaskedarray_to_indexedoptionarray_64(
-      index.ptr().get(),
-      mask_.ptr().get(),
-      mask_.offset(),
+    struct Error err = kernel::BitMaskedArray_to_IndexedOptionArray64(
+      kernel::lib::cpu,   // DERIVE
+      index.data(),
+      mask_.data(),
       mask_.length(),
       valid_when_,
       lsb_order_);
@@ -315,7 +413,8 @@ namespace awkward {
         util::handle_error(
           failure("content and its identities must have the same length",
                   kSliceNone,
-                  kSliceNone),
+                  kSliceNone,
+                  FILENAME_C(__LINE__)),
           classname(),
           identities_.get());
       }
@@ -328,10 +427,10 @@ namespace awkward {
                                          content_.get()->length());
         Identities32* rawsubidentities =
           reinterpret_cast<Identities32*>(subidentities.get());
-        struct Error err = awkward_identities32_extend(
-          rawsubidentities->ptr().get(),
-          rawidentities->ptr().get(),
-          rawidentities->offset(),
+        struct Error err = kernel::Identities_extend<int32_t>(
+          kernel::lib::cpu,   // DERIVE
+          rawsubidentities->data(),
+          rawidentities->data(),
           rawidentities->length(),
           content_.get()->length());
         util::handle_error(err, classname(), identities_.get());
@@ -346,17 +445,19 @@ namespace awkward {
                                          content_.get()->length());
         Identities64* rawsubidentities =
           reinterpret_cast<Identities64*>(subidentities.get());
-        struct Error err = awkward_identities64_extend(
-          rawsubidentities->ptr().get(),
-          rawidentities->ptr().get(),
-          rawidentities->offset(),
+        struct Error err = kernel::Identities_extend<int64_t>(
+          kernel::lib::cpu,   // DERIVE
+          rawsubidentities->data(),
+          rawidentities->data(),
           rawidentities->length(),
           content_.get()->length());
         util::handle_error(err, classname(), identities_.get());
         content_.get()->setidentities(subidentities);
       }
       else {
-        throw std::runtime_error("unrecognized Identities specialization");
+        throw std::runtime_error(
+          std::string("unrecognized Identities specialization")
+          + FILENAME(__LINE__));
       }
     }
     identities_ = identities;
@@ -372,8 +473,10 @@ namespace awkward {
                                        length());
       Identities32* rawidentities =
         reinterpret_cast<Identities32*>(newidentities.get());
-      struct Error err =
-        awkward_new_identities32(rawidentities->ptr().get(), length());
+      struct Error err = kernel::new_Identities<int32_t>(
+        kernel::lib::cpu,   // DERIVE
+        rawidentities->data(),
+        length());
       util::handle_error(err, classname(), identities_.get());
       setidentities(newidentities);
     }
@@ -385,8 +488,10 @@ namespace awkward {
                                        length());
       Identities64* rawidentities =
         reinterpret_cast<Identities64*>(newidentities.get());
-      struct Error err = awkward_new_identities64(rawidentities->ptr().get(),
-                                                  length());
+      struct Error err = kernel::new_Identities<int64_t>(
+        kernel::lib::cpu,   // DERIVE
+        rawidentities->data(),
+        length());
       util::handle_error(err, classname(), identities_.get());
       setidentities(newidentities);
     }
@@ -401,20 +506,21 @@ namespace awkward {
   BitMaskedArray::form(bool materialize) const {
     return std::make_shared<BitMaskedForm>(identities_.get() != nullptr,
                                            parameters_,
+                                           FormKey(nullptr),
                                            mask_.form(),
                                            content_.get()->form(materialize),
                                            valid_when_,
                                            lsb_order_);
   }
 
-  bool
-  BitMaskedArray::has_virtual_form() const {
-    return content_.get()->has_virtual_form();
+  kernel::lib
+  BitMaskedArray::kernels() const {
+    return kernels_compare(mask_.ptr_lib(), content_);
   }
 
-  bool
-  BitMaskedArray::has_virtual_length() const {
-    return content_.get()->has_virtual_length();
+  void
+  BitMaskedArray::caches(std::vector<ArrayCachePtr>& out) const {
+    content_.get()->caches(out);
   }
 
   const std::string
@@ -507,7 +613,10 @@ namespace awkward {
     if (identities_.get() != nullptr  &&
         identities_.get()->length() < length()) {
       util::handle_error(
-        failure("len(identities) < len(array)", kSliceNone, kSliceNone),
+        failure("len(identities) < len(array)",
+                kSliceNone,
+                kSliceNone,
+                FILENAME_C(__LINE__)),
         identities_.get()->classname(),
         nullptr);
     }
@@ -526,7 +635,7 @@ namespace awkward {
     }
     if (!(0 <= regular_at  &&  regular_at < length())) {
       util::handle_error(
-        failure("index out of range", kSliceNone, at),
+        failure("index out of range", kSliceNone, at, FILENAME_C(__LINE__)),
         classname(),
         identities_.get());
     }
@@ -553,12 +662,12 @@ namespace awkward {
   BitMaskedArray::getitem_range(int64_t start, int64_t stop) const {
     int64_t regular_start = start;
     int64_t regular_stop = stop;
-    awkward_regularize_rangeslice(&regular_start, &regular_stop,
+    kernel::regularize_rangeslice(&regular_start, &regular_stop,
       true, start != Slice::none(), stop != Slice::none(), length());
     if (identities_.get() != nullptr  &&
         regular_stop > identities_.get()->length()) {
       util::handle_error(
-        failure("index out of range", kSliceNone, stop),
+        failure("index out of range", kSliceNone, stop, FILENAME_C(__LINE__)),
         identities_.get()->classname(),
         nullptr);
     }
@@ -567,53 +676,57 @@ namespace awkward {
 
   const ContentPtr
   BitMaskedArray::getitem_range_nowrap(int64_t start, int64_t stop) const {
-    int64_t bitstart = start / 8;
-    int64_t remainder = start % 8;
-    if (remainder == 0) {
-      IdentitiesPtr identities(nullptr);
-      if (identities_.get() != nullptr) {
-        identities = identities_.get()->getitem_range_nowrap(start, stop);
-      }
-      int64_t length = stop - start;
-      int64_t bitlength = length / 8;
-      int64_t remainder = length % 8;
-      int64_t bitstop = bitstart + (bitlength + (remainder != 0));
-      return std::make_shared<BitMaskedArray>(
-        identities,
-        parameters_,
-        mask_.getitem_range_nowrap(bitstart, bitstop),
-        content_.get()->getitem_range_nowrap(start, stop),
-        valid_when_,
-        length,
-        lsb_order_);
-    }
-    else {
-      return toByteMaskedArray().get()->getitem_range_nowrap(start, stop);
-    }
+    return toByteMaskedArray().get()->getitem_range_nowrap(start, stop);
   }
 
   const ContentPtr
   BitMaskedArray::getitem_field(const std::string& key) const {
-    return std::make_shared<BitMaskedArray>(
-      identities_,
-      util::Parameters(),
-      mask_,
-      content_.get()->getitem_field(key),
-      valid_when_,
-      length_,
-      lsb_order_);
+    BitMaskedArray step1(identities_,
+                         util::Parameters(),
+                         mask_,
+                         content_.get()->getitem_field(key),
+                         valid_when_,
+                         length_,
+                         lsb_order_);
+    return step1.simplify_optiontype();
+  }
+
+  const ContentPtr
+  BitMaskedArray::getitem_field(const std::string& key,
+                                const Slice& only_fields) const {
+    BitMaskedArray step1(identities_,
+                         util::Parameters(),
+                         mask_,
+                         content_.get()->getitem_field(key, only_fields),
+                         valid_when_,
+                         length_,
+                         lsb_order_);
+    return step1.simplify_optiontype();
   }
 
   const ContentPtr
   BitMaskedArray::getitem_fields(const std::vector<std::string>& keys) const {
-    return std::make_shared<BitMaskedArray>(
-      identities_,
-      util::Parameters(),
-      mask_,
-      content_.get()->getitem_fields(keys),
-      valid_when_,
-      length_,
-      lsb_order_);
+    BitMaskedArray step1(identities_,
+                         util::Parameters(),
+                         mask_,
+                         content_.get()->getitem_fields(keys),
+                         valid_when_,
+                         length_,
+                         lsb_order_);
+    return step1.simplify_optiontype();
+  }
+
+  const ContentPtr
+  BitMaskedArray::getitem_fields(const std::vector<std::string>& keys,
+                                 const Slice& only_fields) const {
+    BitMaskedArray step1(identities_,
+                         util::Parameters(),
+                         mask_,
+                         content_.get()->getitem_fields(keys, only_fields),
+                         valid_when_,
+                         length_,
+                         lsb_order_);
+    return step1.simplify_optiontype();
   }
 
   const ContentPtr
@@ -624,8 +737,31 @@ namespace awkward {
   }
 
   const ContentPtr
-  BitMaskedArray::carry(const Index64& carry) const {
-    return toByteMaskedArray().get()->carry(carry);
+  BitMaskedArray::carry(const Index64& carry, bool allow_lazy) const {
+    if (carry.iscontiguous()) {
+      if (carry.length() == length()) {
+        return shallow_copy();
+      }
+      else {
+        return getitem_range_nowrap(0, carry.length());
+      }
+    }
+    return toByteMaskedArray().get()->carry(carry, allow_lazy);
+  }
+
+  int64_t
+  BitMaskedArray::purelist_depth() const {
+    return content_.get()->purelist_depth();
+  }
+
+  const std::pair<int64_t, int64_t>
+  BitMaskedArray::minmax_depth() const {
+    return content_.get()->minmax_depth();
+  }
+
+  const std::pair<bool, int64_t>
+  BitMaskedArray::branch_depth() const {
+    return content_.get()->branch_depth();
   }
 
   int64_t
@@ -655,7 +791,23 @@ namespace awkward {
 
   const std::string
   BitMaskedArray::validityerror(const std::string& path) const {
-    return content_.get()->validityerror(path + std::string(".content"));
+    const std::string paramcheck = validityerror_parameters(path);
+    if (paramcheck != std::string("")) {
+      return paramcheck;
+    }
+    if (mask_.length() * 8 < length_) {
+      return (std::string("at ") + path + std::string(" (") + classname()
+              + std::string("): ") + std::string("len(mask) * 8 < length")
+              + FILENAME(__LINE__));
+    }
+    else if (content_.get()->length() < length_) {
+      return (std::string("at ") + path + std::string(" (") + classname()
+              + std::string("): ") + std::string("len(content) < length")
+              + FILENAME(__LINE__));
+    }
+    else {
+      return content_.get()->validityerror(path + std::string(".content"));
+    }
   }
 
   const ContentPtr
@@ -679,7 +831,7 @@ namespace awkward {
       return mergeable(raw->array(), mergebool);
     }
 
-    if (!parameters_equal(other.get()->parameters())) {
+    if (!parameters_equal(other.get()->parameters(), false)) {
       return false;
     }
 
@@ -727,17 +879,43 @@ namespace awkward {
     }
   }
 
-  const ContentPtr
-  BitMaskedArray::reverse_merge(const ContentPtr& other) const {
-    ContentPtr indexedoptionarray = toIndexedOptionArray64();
-    IndexedOptionArray64* raw =
-      dynamic_cast<IndexedOptionArray64*>(indexedoptionarray.get());
-    return raw->reverse_merge(other);
+  bool
+  BitMaskedArray::referentially_equal(const ContentPtr& other) const {
+    if (identities_.get() == nullptr  &&  other.get()->identities().get() != nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() == nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() != nullptr) {
+      if (!identities_.get()->referentially_equal(other->identities())) {
+        return false;
+      }
+    }
+    if (BitMaskedArray* raw = dynamic_cast<BitMaskedArray*>(other.get())) {
+      return mask_.referentially_equal(raw->mask())  &&
+             valid_when_ == raw->valid_when()  &&
+             length_ == raw->length()  &&
+             lsb_order_ == raw->lsb_order()  &&
+             parameters_ == raw->parameters()  &&
+             content_.get()->referentially_equal(raw->content());
+    }
+    else {
+      return false;
+    }
   }
 
   const ContentPtr
-  BitMaskedArray::merge(const ContentPtr& other) const {
-    return toIndexedOptionArray64().get()->merge(other);
+  BitMaskedArray::reverse_merge(const ContentPtr& other) const {
+    return toIndexedOptionArray64().get()->reverse_merge(other);
+  }
+
+  const ContentPtr
+  BitMaskedArray::mergemany(const ContentPtrVec& others) const {
+    if (others.empty()) {
+      return shallow_copy();
+    }
+    return toIndexedOptionArray64().get()->mergemany(others);
   }
 
   const SliceItemPtr
@@ -766,6 +944,7 @@ namespace awkward {
   BitMaskedArray::reduce_next(const Reducer& reducer,
                               int64_t negaxis,
                               const Index64& starts,
+                              const Index64& shifts,
                               const Index64& parents,
                               int64_t outlength,
                               bool mask,
@@ -773,6 +952,7 @@ namespace awkward {
     return toByteMaskedArray().get()->reduce_next(reducer,
                                                   negaxis,
                                                   starts,
+                                                  shifts,
                                                   parents,
                                                   outlength,
                                                   mask,
@@ -799,12 +979,47 @@ namespace awkward {
                                                    depth);
   }
 
+  const ContentPtr
+  BitMaskedArray::sort_next(int64_t negaxis,
+                            const Index64& starts,
+                            const Index64& parents,
+                            int64_t outlength,
+                            bool ascending,
+                            bool stable,
+                            bool keepdims) const {
+    return toByteMaskedArray().get()->sort_next(negaxis,
+                                                starts,
+                                                parents,
+                                                outlength,
+                                                ascending,
+                                                stable,
+                                                keepdims);
+  }
+
+  const ContentPtr
+  BitMaskedArray::argsort_next(int64_t negaxis,
+                               const Index64& starts,
+                               const Index64& parents,
+                               int64_t outlength,
+                               bool ascending,
+                               bool stable,
+                               bool keepdims) const {
+    return toByteMaskedArray().get()->argsort_next(negaxis,
+                                                   starts,
+                                                   parents,
+                                                   outlength,
+                                                   ascending,
+                                                   stable,
+                                                   keepdims);
+  }
+
   const ContentPtr
   BitMaskedArray::getitem_next(const SliceAt& at,
                                const Slice& tail,
                                const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: BitMaskedArraygetitem_next(at)");
+      std::string("undefined operation: BitMaskedArraygetitem_next(at)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -812,7 +1027,8 @@ namespace awkward {
                                const Slice& tail,
                                const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: BitMaskedArraygetitem_next(range)");
+      std::string("undefined operation: BitMaskedArraygetitem_next(range)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -820,7 +1036,8 @@ namespace awkward {
                                const Slice& tail,
                                const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: BitMaskedArraygetitem_next(array)");
+      std::string("undefined operation: BitMaskedArraygetitem_next(array)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -828,7 +1045,8 @@ namespace awkward {
                                const Slice& tail,
                                const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: BitMaskedArraygetitem_next(jagged)");
+      std::string("undefined operation: BitMaskedArraygetitem_next(jagged)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -864,4 +1082,65 @@ namespace awkward {
                                                           tail);
   }
 
+  const ContentPtr
+  BitMaskedArray::getitem_next_jagged(const Index64& slicestarts,
+                                      const Index64& slicestops,
+                                      const SliceVarNewAxis& slicecontent,
+                                      const Slice& tail) const {
+    return toByteMaskedArray().get()->getitem_next_jagged(slicestarts,
+                                                          slicestops,
+                                                          slicecontent,
+                                                          tail);
+  }
+
+  const ContentPtr
+  BitMaskedArray::getitem_next(const SliceVarNewAxis& varnewaxis,
+                               const Slice& tail,
+                               const Index64& advanced) const {
+    SliceJagged64 jagged = content_.get()->varaxis_to_jagged(varnewaxis);
+    return getitem_next(jagged, tail, advanced);
+  }
+
+  const SliceJagged64
+  BitMaskedArray::varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const {
+    return content_.get()->varaxis_to_jagged(varnewaxis);
+  }
+
+  const ContentPtr
+  BitMaskedArray::copy_to(kernel::lib ptr_lib) const {
+    IndexU8 mask = mask_.copy_to(ptr_lib);
+    ContentPtr content = content_.get()->copy_to(ptr_lib);
+    IdentitiesPtr identities(nullptr);
+    if (identities_.get() != nullptr) {
+      identities = identities_.get()->copy_to(ptr_lib);
+    }
+    return std::make_shared<BitMaskedArray>(identities,
+                                            parameters_,
+                                            mask,
+                                            content,
+                                            valid_when_,
+                                            length_,
+                                            lsb_order_);
+  }
+
+  const ContentPtr
+  BitMaskedArray::numbers_to_type(const std::string& name) const {
+    return toByteMaskedArray().get()->numbers_to_type(name);
+  }
+
+  bool
+  BitMaskedArray::is_unique() const {
+    return toByteMaskedArray().get()->is_unique();
+  }
+
+  const ContentPtr
+  BitMaskedArray::unique() const {
+    return toByteMaskedArray().get()->unique();
+  }
+
+  bool
+  BitMaskedArray::is_subrange_equal(const Index64& start, const Index64& stop) const {
+    return toByteMaskedArray().get()->is_subrange_equal(start, stop);
+  }
+
 }
