@@ -1,15 +1,20 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
+
+#define FILENAME(line) FILENAME_FOR_EXCEPTIONS("src/libawkward/array/EmptyArray.cpp", line)
+#define FILENAME_C(line) FILENAME_FOR_EXCEPTIONS_C("src/libawkward/array/EmptyArray.cpp", line)
 
 #include <iomanip>
 #include <sstream>
 #include <stdexcept>
 
-#include "awkward/cpu-kernels/operations.h"
+#include "awkward/kernels.h"
 #include "awkward/type/UnknownType.h"
-#include "awkward/type/ArrayType.h"
+#include "awkward/Reducer.h"
+#include "awkward/io/json.h"
 #include "awkward/array/IndexedArray.h"
 #include "awkward/array/NumpyArray.h"
 #include "awkward/array/RegularArray.h"
+#include "awkward/array/VirtualArray.h"
 
 #include "awkward/array/EmptyArray.h"
 
@@ -17,8 +22,9 @@ namespace awkward {
   ////////// EmptyForm
 
   EmptyForm::EmptyForm(bool has_identities,
-                       const util::Parameters& parameters)
-      : Form(has_identities, parameters) { }
+                       const util::Parameters& parameters,
+                       const FormKey& form_key)
+      : Form(has_identities, parameters, form_key) { }
 
   const TypePtr
   EmptyForm::type(const util::TypeStrs& typestrs) const {
@@ -34,13 +40,22 @@ namespace awkward {
     builder.string("EmptyArray");
     identities_tojson(builder, verbose);
     parameters_tojson(builder, verbose);
+    form_key_tojson(builder, verbose);
     builder.endrecord();
   }
 
   const FormPtr
   EmptyForm::shallow_copy() const {
     return std::make_shared<EmptyForm>(has_identities_,
-                                       parameters_);
+                                       parameters_,
+                                       form_key_);
+  }
+
+  const FormPtr
+  EmptyForm::with_form_key(const FormKey& form_key) const {
+    return std::make_shared<EmptyForm>(has_identities_,
+                                       parameters_,
+                                       form_key);
   }
 
   const std::string
@@ -58,6 +73,11 @@ namespace awkward {
     return 1;
   }
 
+  bool
+  EmptyForm::dimension_optiontype() const {
+    return false;
+  }
+
   const std::pair<int64_t, int64_t>
   EmptyForm::minmax_depth() const {
     return std::pair<int64_t, int64_t>(1, 1);
@@ -76,15 +96,17 @@ namespace awkward {
   int64_t
   EmptyForm::fieldindex(const std::string& key) const {
     throw std::invalid_argument(
-      std::string("key ") + util::quote(key, true)
-      + std::string(" does not exist (data might not be records)"));
+      std::string("key ") + util::quote(key)
+      + std::string(" does not exist (data might not be records)")
+      + FILENAME(__LINE__));
   }
 
   const std::string
   EmptyForm::key(int64_t fieldindex) const {
     throw std::invalid_argument(
       std::string("fieldindex \"") + std::to_string(fieldindex)
-      + std::string("\" does not exist (data might not be records)"));
+      + std::string("\" does not exist (data might not be records)")
+      + FILENAME(__LINE__));
   }
 
   bool
@@ -100,13 +122,31 @@ namespace awkward {
   bool
   EmptyForm::equal(const FormPtr& other,
                    bool check_identities,
-                   bool check_parameters) const {
+                   bool check_parameters,
+                   bool check_form_key,
+                   bool compatibility_check) const {
+    if (compatibility_check) {
+      if (VirtualForm* raw = dynamic_cast<VirtualForm*>(other.get())) {
+        if (raw->form().get() != nullptr) {
+          return equal(raw->form(),
+                       check_identities,
+                       check_parameters,
+                       check_form_key,
+                       compatibility_check);
+        }
+      }
+    }
+
     if (check_identities  &&
         has_identities_ != other.get()->has_identities()) {
       return false;
     }
     if (check_parameters  &&
-        !util::parameters_equal(parameters_, other.get()->parameters())) {
+        !util::parameters_equal(parameters_, other.get()->parameters(), false)) {
+      return false;
+    }
+    if (check_form_key  &&
+        !form_key_equals(other.get()->form_key())) {
       return false;
     }
     if (EmptyForm* t = dynamic_cast<EmptyForm*>(other.get())) {
@@ -117,6 +157,19 @@ namespace awkward {
     }
   }
 
+  const FormPtr
+  EmptyForm::getitem_field(const std::string& key) const {
+    throw std::invalid_argument(
+      std::string("key ") + util::quote(key)
+      + std::string(" does not exist (data might not be records)"));
+  }
+
+  const FormPtr
+  EmptyForm::getitem_fields(const std::vector<std::string>& keys) const {
+    throw std::invalid_argument(
+      std::string("requested keys do not exist (data might not be records)"));
+  }
+
   ////////// EmptyArray
 
   EmptyArray::EmptyArray(const IdentitiesPtr& identities,
@@ -124,8 +177,10 @@ namespace awkward {
       : Content(identities, parameters) { }
 
   const std::shared_ptr<NumpyArray>
-  EmptyArray::toNumpyArray(const std::string& format, ssize_t itemsize) const {
-    std::shared_ptr<void> ptr(new uint8_t[0], util::array_deleter<uint8_t>());
+  EmptyArray::toNumpyArray(const std::string& format,
+                           ssize_t itemsize,
+                           util::dtype dtype) const {
+    std::shared_ptr<void> ptr = kernel::malloc<uint8_t>(kernel::lib::cpu, 0);
     std::vector<ssize_t> shape({ 0 });
     std::vector<ssize_t> strides({ itemsize });
     return std::make_shared<NumpyArray>(identities_,
@@ -135,7 +190,9 @@ namespace awkward {
                                         strides,
                                         0,
                                         itemsize,
-                                        format);
+                                        format,
+                                        dtype,
+                                        kernel::lib::cpu);
   }
 
   const std::string
@@ -150,7 +207,8 @@ namespace awkward {
       util::handle_error(
         failure("content and its identities must have the same length",
                 kSliceNone,
-                kSliceNone),
+                kSliceNone,
+                FILENAME_C(__LINE__)),
         classname(),
         identities_.get());
     }
@@ -168,18 +226,17 @@ namespace awkward {
   const FormPtr
   EmptyArray::form(bool materialize) const {
     return std::make_shared<EmptyForm>(identities_.get() != nullptr,
-                                       parameters_);
+                                       parameters_,
+                                       FormKey(nullptr));
   }
 
-  bool
-  EmptyArray::has_virtual_form() const {
-    return false;
+  kernel::lib
+  EmptyArray::kernels() const {
+    return kernel::lib::cpu;
   }
 
-  bool
-  EmptyArray::has_virtual_length() const {
-    return false;
-  }
+  void
+  EmptyArray::caches(std::vector<ArrayCachePtr>& out) const { }
 
   const std::string
   EmptyArray::tostring_part(const std::string& indent,
@@ -254,7 +311,10 @@ namespace awkward {
   const ContentPtr
   EmptyArray::getitem_at(int64_t at) const {
     util::handle_error(
-      failure("index out of range", kSliceNone, at),
+      failure("index out of range",
+              kSliceNone,
+              at,
+              FILENAME_C(__LINE__)),
       classname(),
       identities_.get());
     return ContentPtr(nullptr);  // make Windows compiler happy
@@ -263,7 +323,7 @@ namespace awkward {
   const ContentPtr
   EmptyArray::getitem_at_nowrap(int64_t at) const {
     util::handle_error(
-      failure("index out of range", kSliceNone, at),
+      failure("index out of range", kSliceNone, at, FILENAME_C(__LINE__)),
       classname(),
       identities_.get());
     return ContentPtr(nullptr);  // make Windows compiler happy
@@ -283,19 +343,57 @@ namespace awkward {
   EmptyArray::getitem_field(const std::string& key) const {
     throw std::invalid_argument(
       std::string("cannot slice ") + classname()
-      + std::string(" by field name"));
+      + std::string(" by field name") + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  EmptyArray::getitem_field(const std::string& key,
+                            const Slice& only_fields) const {
+    throw std::invalid_argument(
+      std::string("cannot slice ") + classname()
+      + std::string(" by field name") + FILENAME(__LINE__));
   }
 
   const ContentPtr
   EmptyArray::getitem_fields(const std::vector<std::string>& keys) const {
     throw std::invalid_argument(
       std::string("cannot slice ") + classname()
-      + std::string(" by field names"));
+      + std::string(" by field names") + FILENAME(__LINE__));
   }
 
   const ContentPtr
-  EmptyArray::carry(const Index64& carry) const {
-    return shallow_copy();
+  EmptyArray::getitem_fields(const std::vector<std::string>& keys,
+                             const Slice& only_fields) const {
+    throw std::invalid_argument(
+      std::string("cannot slice ") + classname()
+      + std::string(" by field names") + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  EmptyArray::carry(const Index64& carry, bool allow_lazy) const {
+    if (carry.length() == 0) {
+      return shallow_copy();
+    }
+    else {
+    throw std::invalid_argument(
+      std::string("cannot extract ") + std::to_string(carry.length()) +
+      std::string(" elements from ") + classname() + FILENAME(__LINE__));
+    }
+  }
+
+  int64_t
+  EmptyArray::purelist_depth() const {
+    return 1;
+  }
+
+  const std::pair<int64_t, int64_t>
+  EmptyArray::minmax_depth() const {
+    return std::pair<int64_t, int64_t>(1, 1);
+  }
+
+  const std::pair<bool, int64_t>
+  EmptyArray::branch_depth() const {
+    return std::pair<bool, int64_t>(false, 1);
   }
 
   int64_t
@@ -306,15 +404,17 @@ namespace awkward {
   int64_t
   EmptyArray::fieldindex(const std::string& key) const {
     throw std::invalid_argument(
-      std::string("key ") + util::quote(key, true)
-      + std::string(" does not exist (data might not be records)"));
+      std::string("key ") + util::quote(key)
+      + std::string(" does not exist (data might not be records)")
+      + FILENAME(__LINE__));
   }
 
   const std::string
   EmptyArray::key(int64_t fieldindex) const {
     throw std::invalid_argument(
       std::string("fieldindex \"") + std::to_string(fieldindex)
-      + std::string("\" does not exist (data might not be records)"));
+      + std::string("\" does not exist (data might not be records)")
+      + FILENAME(__LINE__));
   }
 
   bool
@@ -329,6 +429,10 @@ namespace awkward {
 
   const std::string
   EmptyArray::validityerror(const std::string& path) const {
+    const std::string paramcheck = validityerror_parameters(path);
+    if (paramcheck != std::string("")) {
+      return paramcheck;
+    }
     return std::string();
   }
 
@@ -339,8 +443,8 @@ namespace awkward {
 
   const ContentPtr
   EmptyArray::num(int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       Index64 out(1);
       out.setitem_at_nowrap(0, length());
       return NumpyArray(out).getitem_at_nowrap(0);
@@ -352,9 +456,10 @@ namespace awkward {
 
   const std::pair<Index64, ContentPtr>
   EmptyArray::offsets_and_flattened(int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
-      throw std::invalid_argument("axis=0 not allowed for flatten");
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
+      throw std::invalid_argument(
+        std::string("axis=0 not allowed for flatten") + FILENAME(__LINE__));
     }
     else {
       Index64 offsets(1);
@@ -367,12 +472,47 @@ namespace awkward {
 
   bool
   EmptyArray::mergeable(const ContentPtr& other, bool mergebool) const {
+    if (!parameters_equal(other.get()->parameters(), false)) {
+      return false;
+    }
     return true;
   }
 
+  bool
+  EmptyArray::referentially_equal(const ContentPtr& other) const {
+    if (identities_.get() == nullptr  &&  other.get()->identities().get() != nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() == nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() != nullptr) {
+      if (!identities_.get()->referentially_equal(other->identities())) {
+        return false;
+      }
+    }
+    if (EmptyArray* raw = dynamic_cast<EmptyArray*>(other.get())) {
+      return parameters_ == raw->parameters();
+    }
+    else {
+      return false;
+    }
+  }
+
   const ContentPtr
-  EmptyArray::merge(const ContentPtr& other) const {
-    return other;
+  EmptyArray::mergemany(const ContentPtrVec& others) const {
+    if (others.empty()) {
+      return shallow_copy();
+    }
+
+    else if (others.size() == 1) {
+      return others[0];
+    }
+
+    else {
+      ContentPtrVec tail_others(others.begin() + 1, others.end());
+      return others[0].get()->mergemany(tail_others);
+    }
   }
 
   const SliceItemPtr
@@ -391,12 +531,13 @@ namespace awkward {
 
   const ContentPtr
   EmptyArray::rpad(int64_t target, int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis != depth) {
-      throw std::invalid_argument("axis exceeds the depth of this array");
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis != depth) {
+      throw std::invalid_argument(
+        std::string("axis exceeds the depth of this array") + FILENAME(__LINE__));
     }
     else {
-      return rpad_and_clip(target, axis, depth);
+      return rpad_and_clip(target, posaxis, depth);
     }
   }
 
@@ -404,9 +545,10 @@ namespace awkward {
   EmptyArray::rpad_and_clip(int64_t target,
                             int64_t axis,
                             int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis != depth) {
-      throw std::invalid_argument("axis exceeds the depth of this array");
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis != depth) {
+      throw std::invalid_argument(
+        std::string("axis exceeds the depth of this array") + FILENAME(__LINE__));
     }
     else {
       return rpad_axis0(target, true);
@@ -417,15 +559,19 @@ namespace awkward {
   EmptyArray::reduce_next(const Reducer& reducer,
                           int64_t negaxis,
                           const Index64& starts,
+                          const Index64& shifts,
                           const Index64& parents,
                           int64_t outlength,
                           bool mask,
                           bool keepdims) const {
-    ContentPtr asnumpy = toNumpyArray(reducer.preferred_type(),
-                                      reducer.preferred_typesize());
+    util::dtype dtype = reducer.preferred_dtype();
+    std::string format = util::dtype_to_format(dtype);
+    int64_t itemsize = util::dtype_to_itemsize(dtype);
+    ContentPtr asnumpy = toNumpyArray(format, itemsize, dtype);
     return asnumpy.get()->reduce_next(reducer,
                                       negaxis,
                                       starts,
+                                      shifts,
                                       parents,
                                       outlength,
                                       mask,
@@ -445,17 +591,58 @@ namespace awkward {
                            int64_t axis,
                            int64_t depth) const {
     if (n < 1) {
-      throw std::invalid_argument("in combinations, 'n' must be at least 1");
+      throw std::invalid_argument(
+        std::string("in combinations, 'n' must be at least 1") + FILENAME(__LINE__));
     }
     return std::make_shared<EmptyArray>(identities_, util::Parameters());
   }
 
+  const ContentPtr
+  EmptyArray::sort_next(int64_t negaxis,
+                        const Index64& starts,
+                        const Index64& parents,
+                        int64_t outlength,
+                        bool ascending,
+                        bool stable,
+                        bool keepdims) const {
+    ContentPtr asnumpy = toNumpyArray("d", 8, util::dtype::float64);
+    return asnumpy.get()->sort_next(negaxis,
+                                    starts,
+                                    parents,
+                                    outlength,
+                                    ascending,
+                                    stable,
+                                    keepdims);
+  }
+
+  const ContentPtr
+  EmptyArray::argsort_next(int64_t negaxis,
+                           const Index64& starts,
+                           const Index64& parents,
+                           int64_t outlength,
+                           bool ascending,
+                           bool stable,
+                           bool keepdims) const {
+    ContentPtr asnumpy = toNumpyArray("d", 8, util::dtype::float64);
+    ContentPtr out = asnumpy.get()->argsort_next(negaxis,
+                                                 starts,
+                                                 parents,
+                                                 outlength,
+                                                 ascending,
+                                                 stable,
+                                                 keepdims);
+    return out;
+  }
+
   const ContentPtr
   EmptyArray::getitem_next(const SliceAt& at,
                            const Slice& tail,
                            const Index64& advanced) const {
     util::handle_error(
-      failure("too many dimensions in slice", kSliceNone, kSliceNone),
+      failure("too many dimensions in slice",
+              kSliceNone,
+              kSliceNone,
+              FILENAME_C(__LINE__)),
       classname(),
       identities_.get());
     return ContentPtr(nullptr);  // make Windows compiler happy
@@ -466,7 +653,10 @@ namespace awkward {
                            const Slice& tail,
                            const Index64& advanced) const {
     util::handle_error(
-      failure("too many dimensions in slice", kSliceNone, kSliceNone),
+      failure("too many dimensions in slice",
+              kSliceNone,
+              kSliceNone,
+              FILENAME_C(__LINE__)),
       classname(),
       identities_.get());
     return ContentPtr(nullptr);  // make Windows compiler happy
@@ -477,7 +667,10 @@ namespace awkward {
                            const Slice& tail,
                            const Index64& advanced) const {
     util::handle_error(
-      failure("too many dimensions in slice", kSliceNone, kSliceNone),
+      failure("too many dimensions in slice",
+              kSliceNone,
+              kSliceNone,
+              FILENAME_C(__LINE__)),
       classname(),
       identities_.get());
     return ContentPtr(nullptr);  // make Windows compiler happy
@@ -489,7 +682,8 @@ namespace awkward {
                            const Index64& advanced) const {
     throw std::invalid_argument(
       std::string("cannot slice ") + classname()
-      + std::string(" by a field name because it has no fields"));
+      + std::string(" by a field name because it has no fields")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -498,18 +692,21 @@ namespace awkward {
                            const Index64& advanced) const {
     throw std::invalid_argument(
       std::string("cannot slice ") + classname()
-      + std::string(" by field names because it has no fields"));
+      + std::string(" by field names because it has no fields")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
   EmptyArray::getitem_next(const SliceJagged64& jagged,
                            const Slice& tail,
                            const Index64& advanced) const {
-    if (advanced.length() != 0) {
+    if (!advanced.is_empty_advanced()) {
       throw std::invalid_argument(
-        "cannot mix jagged slice with NumPy-style advanced indexing");
+        std::string("cannot mix jagged slice with NumPy-style advanced indexing")
+        + FILENAME(__LINE__));
     }
-    throw std::runtime_error("FIXME: EmptyArray::getitem_next(jagged)");
+    throw std::runtime_error(
+      std::string("FIXME: EmptyArray::getitem_next(jagged)") + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -518,7 +715,8 @@ namespace awkward {
                                   const SliceArray64& slicecontent,
                                   const Slice& tail) const {
     throw std::runtime_error(
-      "undefined operation: EmptyArray::getitem_next_jagged(array)");
+      std::string("undefined operation: EmptyArray::getitem_next_jagged(array)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -527,7 +725,8 @@ namespace awkward {
                                   const SliceMissing64& slicecontent,
                                   const Slice& tail) const {
     throw std::runtime_error(
-      "undefined operation: EmptyArray::getitem_next_jagged(missing)");
+      std::string("undefined operation: EmptyArray::getitem_next_jagged(missing)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -536,7 +735,75 @@ namespace awkward {
                                   const SliceJagged64& slicecontent,
                                   const Slice& tail) const {
     throw std::runtime_error(
-      "undefined operation: EmptyArray::getitem_next_jagged(jagged)");
+      std::string("undefined operation: EmptyArray::getitem_next_jagged(jagged)")
+      + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  EmptyArray::getitem_next_jagged(const Index64& slicestarts,
+                                  const Index64& slicestops,
+                                  const SliceVarNewAxis& slicecontent,
+                                  const Slice& tail) const {
+    throw std::runtime_error(
+      std::string("undefined operation: EmptyArray::getitem_next_jagged(varnewaxis)")
+      + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  EmptyArray::getitem_next(const SliceVarNewAxis& varnewaxis,
+                           const Slice& tail,
+                           const Index64& advanced) const {
+    util::handle_error(
+      failure("too many dimensions in slice",
+              kSliceNone,
+              kSliceNone,
+              FILENAME_C(__LINE__)),
+      classname(),
+      identities_.get());
+    return ContentPtr(nullptr);  // make Windows compiler happy
+  }
+
+  const SliceJagged64
+  EmptyArray::varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const {
+    util::handle_error(
+      failure("too many dimensions in slice",
+              kSliceNone,
+              kSliceNone,
+              FILENAME_C(__LINE__)),
+      classname(),
+      identities_.get());
+    return SliceJagged64(Index64(0), SliceItemPtr(nullptr));  // make Windows compiler happy
+  }
+
+  const ContentPtr
+  EmptyArray::copy_to(kernel::lib ptr_lib) const {
+    IdentitiesPtr identities(nullptr);
+    if (identities_.get() != nullptr) {
+      identities = identities_.get()->copy_to(ptr_lib);
+    }
+    return std::make_shared<EmptyArray>(identities,
+                                        parameters_);
+  }
+
+  const ContentPtr
+  EmptyArray::numbers_to_type(const std::string& name) const {
+    return std::make_shared<EmptyArray>(identities(),
+                                        parameters());
+  }
+
+  bool
+  EmptyArray::is_unique() const {
+    return true;
+  }
+
+  const ContentPtr
+  EmptyArray::unique() const {
+    return shallow_copy();
+  }
+
+  bool
+  EmptyArray::is_subrange_equal(const Index64& start, const Index64& stop) const {
+    return false;
   }
 
 }
