@@ -1,13 +1,16 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
+
+#define FILENAME(line) FILENAME_FOR_EXCEPTIONS("src/libawkward/array/Record.cpp", line)
+#define FILENAME_C(line) FILENAME_FOR_EXCEPTIONS_C("src/libawkward/array/Record.cpp", line)
 
 #include <sstream>
+#include <memory>
 
-#include "awkward/cpu-kernels/identities.h"
-#include "awkward/cpu-kernels/getitem.h"
+#include "awkward/kernels.h"
 #include "awkward/type/RecordType.h"
-#include "awkward/type/ArrayType.h"
-
+#include "awkward/Reducer.h"
 #include "awkward/array/Record.h"
+#include "awkward/io/json.h"
 
 namespace awkward {
   Record::Record(const std::shared_ptr<const RecordArray> array, int64_t at)
@@ -17,7 +20,8 @@ namespace awkward {
     if (!(0 <= at  &&  at < array.get()->length())) {
       throw std::invalid_argument(
         std::string("at=") + std::to_string(at)
-        + std::string(" is out of range for recordarray"));
+        + std::string(" is out of range for recordarray")
+        + FILENAME(__LINE__));
     }
   }
 
@@ -73,12 +77,16 @@ namespace awkward {
 
   void
   Record::setidentities() {
-    throw std::runtime_error("undefined operation: Record::setidentities");
+    throw std::runtime_error(
+      std::string("undefined operation: Record::setidentities")
+      + FILENAME(__LINE__));
   }
 
   void
   Record::setidentities(const IdentitiesPtr& identities) {
-    throw std::runtime_error("undefined operation: Record::setidentities");
+    throw std::runtime_error(
+      std::string("undefined operation: Record::setidentities")
+      + FILENAME(__LINE__));
   }
 
   const TypePtr
@@ -91,14 +99,14 @@ namespace awkward {
     return array_.get()->form(materialize);
   }
 
-  bool
-  Record::has_virtual_form() const {
-    return array_.get()->has_virtual_form();
+  kernel::lib
+  Record::kernels() const {
+    return array_.get()->kernels();
   }
 
-  bool
-  Record::has_virtual_length() const {
-    return array_.get()->has_virtual_length();
+  void
+  Record::caches(std::vector<ArrayCachePtr>& out) const {
+    array_.get()->caches(out);
   }
 
   const std::string
@@ -168,7 +176,8 @@ namespace awkward {
       util::handle_error(
         failure("len(identities) != 1 for scalar Record",
                 kSliceNone,
-                kSliceNone),
+                kSliceNone,
+                FILENAME_C(__LINE__)),
         array_.get()->identities().get()->classname(),
         nullptr);
     }
@@ -176,33 +185,39 @@ namespace awkward {
 
   const ContentPtr
   Record::getitem_nothing() const {
-    throw std::runtime_error("undefined operation: Record::getitem_nothing");
+    throw std::runtime_error(
+      std::string("undefined operation: Record::getitem_nothing")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
   Record::getitem_at(int64_t at) const {
     throw std::invalid_argument(
       std::string("scalar Record can only be sliced by field name (string); "
-                  "try ") + util::quote(std::to_string(at), true));
+                  "try ") + util::quote(std::to_string(at))
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
   Record::getitem_at_nowrap(int64_t at) const {
     throw std::invalid_argument(
       std::string("scalar Record can only be sliced by field name (string); "
-                  "try ") + util::quote(std::to_string(at), true));
+                  "try ") + util::quote(std::to_string(at))
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
   Record::getitem_range(int64_t start, int64_t stop) const {
     throw std::invalid_argument(
-      "scalar Record can only be sliced by field name (string)");
+      std::string("scalar Record can only be sliced by field name (string)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
   Record::getitem_range_nowrap(int64_t start, int64_t stop) const {
     throw std::invalid_argument(
-      "scalar Record can only be sliced by field name (string)");
+      std::string("scalar Record can only be sliced by field name (string)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -210,6 +225,13 @@ namespace awkward {
     return array_.get()->field(key).get()->getitem_at_nowrap(at_);
   }
 
+  const ContentPtr
+  Record::getitem_field(const std::string& key,
+                        const Slice& only_fields) const {
+    ContentPtr recordarray = array_.get()->getitem_field(key, only_fields);
+    return recordarray.get()->getitem_at_nowrap(at_);
+  }
+
   const ContentPtr
   Record::getitem_fields(const std::vector<std::string>& keys) const {
     ContentPtr recordarray = array_.get()->getitem_fields(keys);
@@ -217,8 +239,21 @@ namespace awkward {
   }
 
   const ContentPtr
-  Record::carry(const Index64& carry) const {
-    throw std::runtime_error("undefined operation: Record::carry");
+  Record::getitem_fields(const std::vector<std::string>& keys,
+                         const Slice& only_fields) const {
+    ContentPtr recordarray = array_.get()->getitem_fields(keys, only_fields);
+    return recordarray.get()->getitem_at_nowrap(at_);
+  }
+
+  const ContentPtr
+  Record::carry(const Index64& carry, bool allow_lazy) const {
+    throw std::runtime_error(
+      std::string("undefined operation: Record::carry") + FILENAME(__LINE__));
+  }
+
+  bool
+  Record::dimension_optiontype() const {
+    return false;
   }
 
   int64_t
@@ -275,38 +310,66 @@ namespace awkward {
 
   const ContentPtr
   Record::num(int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       throw std::invalid_argument(
-        "cannot call 'num' with an 'axis' of 0 on a Record");
+        std::string("cannot call 'num' with an 'axis' of 0 on a Record")
+        + FILENAME(__LINE__));
     }
     else {
       ContentPtr singleton = array_.get()->getitem_range_nowrap(at_, at_ + 1);
-      return singleton.get()->num(axis, depth).get()->getitem_at_nowrap(0);
+      return singleton.get()->num(posaxis, depth).get()->getitem_at_nowrap(0);
     }
   }
 
   const std::pair<Index64, ContentPtr>
   Record::offsets_and_flattened(int64_t axis, int64_t depth) const {
     throw std::invalid_argument(
-      "Record cannot be flattened because it is not an array");
+      std::string("Record cannot be flattened because it is not an array")
+      + FILENAME(__LINE__));
   }
 
   bool
   Record::mergeable(const ContentPtr& other, bool mergebool) const {
     throw std::invalid_argument(
-      "Record cannot be merged because it is not an array");
+      std::string("Record cannot be merged because it is not an array")
+      + FILENAME(__LINE__));
+  }
+
+  bool
+  Record::referentially_equal(const ContentPtr& other) const {
+    if (identities_.get() == nullptr  &&  other.get()->identities().get() != nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() == nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() != nullptr) {
+      if (!identities_.get()->referentially_equal(other->identities())) {
+        return false;
+      }
+    }
+    if (Record* raw = dynamic_cast<Record*>(other.get())) {
+      return at_ == raw->at()  &&
+             parameters_ == raw->parameters()  &&
+             array_.get()->referentially_equal(raw->shallow_copy());
+    }
+    else {
+      return false;
+    }
   }
 
   const ContentPtr
-  Record::merge(const ContentPtr& other) const {
+  Record::mergemany(const ContentPtrVec& others) const {
     throw std::invalid_argument(
-      "Record cannot be merged because it is not an array");
+      std::string("Record cannot be merged because it is not an array")
+      + FILENAME(__LINE__));
   }
 
   const SliceItemPtr
   Record::asslice() const {
-    throw std::invalid_argument("cannot use a record as a slice");
+    throw std::invalid_argument(
+      std::string("cannot use a record as a slice") + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -320,19 +383,22 @@ namespace awkward {
   const ContentPtr
   Record::rpad(int64_t length, int64_t axis, int64_t depth) const {
     throw std::invalid_argument(
-      "Record cannot be padded because it is not an array");
+      std::string("Record cannot be padded because it is not an array")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
   Record::rpad_and_clip(int64_t length, int64_t axis, int64_t depth) const {
     throw std::invalid_argument(
-      "Record cannot be padded because it is not an array");
+      std::string("Record cannot be padded because it is not an array")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
   Record::reduce_next(const Reducer& reducer,
                       int64_t negaxis,
                       const Index64& starts,
+                      const Index64& shifts,
                       const Index64& parents,
                       int64_t outlength,
                       bool mask,
@@ -341,6 +407,7 @@ namespace awkward {
     return trimmed.get()->reduce_next(reducer,
                                       negaxis,
                                       starts,
+                                      shifts,
                                       parents,
                                       outlength,
                                       mask,
@@ -349,15 +416,16 @@ namespace awkward {
 
   const ContentPtr
   Record::localindex(int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       throw std::invalid_argument(
-        "cannot call 'localindex' with an 'axis' of 0 on a Record");
+        std::string("cannot call 'localindex' with an 'axis' of 0 on a Record")
+        + FILENAME(__LINE__));
     }
     else {
       ContentPtr singleton = array_.get()->getitem_range_nowrap(at_, at_ + 1);
       return singleton.get()
-             ->localindex(axis, depth).get()
+             ->localindex(posaxis, depth).get()
              ->getitem_at_nowrap(0);
     }
   }
@@ -370,12 +438,14 @@ namespace awkward {
                        int64_t axis,
                        int64_t depth) const {
     if (n < 1) {
-      throw std::invalid_argument("in combinations, 'n' must be at least 1");
-    }
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
       throw std::invalid_argument(
-        "cannot call 'combinations' with an 'axis' of 0 on a Record");
+        std::string("in combinations, 'n' must be at least 1") + FILENAME(__LINE__));
+    }
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
+      throw std::invalid_argument(
+        std::string("cannot call 'combinations' with an 'axis' of 0 on a Record")
+        + FILENAME(__LINE__));
     }
     else {
       ContentPtr singleton = array_.get()->getitem_range_nowrap(at_, at_ + 1);
@@ -383,7 +453,7 @@ namespace awkward {
                                            replacement,
                                            recordlookup,
                                            parameters,
-                                           axis,
+                                           posaxis,
                                            depth).get()->getitem_at_nowrap(0);
     }
   }
@@ -436,13 +506,49 @@ namespace awkward {
     return std::make_shared<Record>(array_.get()->astuple(), at_);
   }
 
+  const ContentPtr
+  Record::sort_next(int64_t negaxis,
+                    const Index64& starts,
+                    const Index64& parents,
+                    int64_t outlength,
+                    bool ascending,
+                    bool stable,
+                    bool keepdims) const {
+    ContentPtr next = array_.get()->getitem_at_nowrap(at_);
+    return next.get()->sort_next(negaxis,
+                                 starts,
+                                 parents,
+                                 outlength,
+                                 ascending,
+                                 stable,
+                                 keepdims);
+  }
+
+  const ContentPtr
+  Record::argsort_next(int64_t negaxis,
+                       const Index64& starts,
+                       const Index64& parents,
+                       int64_t outlength,
+                       bool ascending,
+                       bool stable,
+                       bool keepdims) const {
+    ContentPtr next = array_.get()->getitem_at_nowrap(at_);
+    return next.get()->argsort_next(negaxis,
+                                    starts,
+                                    parents,
+                                    outlength,
+                                    ascending,
+                                    stable,
+                                    keepdims);
+  }
+
   const ContentPtr
   Record::getitem(const Slice& where) const {
     ContentPtr next = array_.get()->getitem_range_nowrap(at_, at_ + 1);
 
     SliceItemPtr nexthead = where.head();
     Slice nexttail = where.tail();
-    Index64 nextadvanced(0);
+    Index64 nextadvanced = Index64::empty_advanced();
     ContentPtr out = next.get()->getitem_next(nexthead,
                                               nexttail,
                                               nextadvanced);
@@ -459,7 +565,9 @@ namespace awkward {
   Record::getitem_next(const SliceAt& at,
                        const Slice& tail,
                        const Index64& advanced) const {
-    throw std::runtime_error("undefined operation: Record::getitem_next(at)");
+    throw std::runtime_error(
+      std::string("undefined operation: Record::getitem_next(at)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -467,7 +575,8 @@ namespace awkward {
                        const Slice& tail,
                        const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next(range)");
+      std::string("undefined operation: Record::getitem_next(range)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -475,7 +584,8 @@ namespace awkward {
                        const Slice& tail,
                        const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next(array)");
+      std::string("undefined operation: Record::getitem_next(array)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -483,7 +593,8 @@ namespace awkward {
                        const Slice& tail,
                        const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next(field)");
+      std::string("undefined operation: Record::getitem_next(field)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -491,7 +602,8 @@ namespace awkward {
                        const Slice& tail,
                        const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next(fields)");
+      std::string("undefined operation: Record::getitem_next(fields)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -499,7 +611,8 @@ namespace awkward {
                        const Slice& tail,
                        const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next(jagged)");
+      std::string("undefined operation: Record::getitem_next(jagged)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -508,7 +621,8 @@ namespace awkward {
                               const SliceArray64& slicecontent,
                               const Slice& tail) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next_jagged(array)");
+      std::string("undefined operation: Record::getitem_next_jagged(array)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -517,7 +631,8 @@ namespace awkward {
                               const SliceMissing64& slicecontent,
                               const Slice& tail) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next_jagged(missing)");
+      std::string("undefined operation: Record::getitem_next_jagged(missing)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -526,7 +641,69 @@ namespace awkward {
                               const SliceJagged64& slicecontent,
                               const Slice& tail) const {
     throw std::runtime_error(
-      "undefined operation: Record::getitem_next_jagged(jagged)");
+      std::string("undefined operation: Record::getitem_next_jagged(jagged)")
+      + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  Record::getitem_next_jagged(const Index64& slicestarts,
+                              const Index64& slicestops,
+                              const SliceVarNewAxis& slicecontent,
+                              const Slice& tail) const {
+    throw std::runtime_error(
+      std::string("undefined operation: Record::getitem_next_jagged(varnewaxis)")
+      + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  Record::getitem_next(const SliceVarNewAxis& varnewaxis,
+                       const Slice& tail,
+                       const Index64& advanced) const {
+    throw std::runtime_error(
+      std::string("undefined operation: Record::getitem_next(varnewaxis)")
+      + FILENAME(__LINE__));
+  }
+
+  const SliceJagged64
+  Record::varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const {
+    throw std::runtime_error(
+      std::string("undefined operation: Record::varaxis_to_jagged(varnewaxis)")
+      + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  Record::copy_to(kernel::lib ptr_lib) const {
+    ContentPtr array = array_.get()->copy_to(ptr_lib);
+    return std::make_shared<Record>(
+             std::dynamic_pointer_cast<RecordArray>(array), at_);
+  }
+
+  const ContentPtr
+  Record::numbers_to_type(const std::string& name) const {
+    ContentPtr out = array_.get()->numbers_to_type(name);
+    return std::make_shared<Record>(
+      std::dynamic_pointer_cast<RecordArray>(out), at_);
+  }
+
+  bool
+  Record::is_unique() const {
+    throw std::runtime_error(
+      std::string("FIXME: operation not yet implemented: Record::is_unique")
+      + FILENAME(__LINE__));
+  }
+
+  const ContentPtr
+  Record::unique() const {
+    throw std::runtime_error(
+      std::string("FIXME: operation not yet implemented: Record::unique")
+      + FILENAME(__LINE__));
+  }
+
+  bool
+  Record::is_subrange_equal(const Index64& start, const Index64& stop) const {
+    throw std::runtime_error(
+      std::string("FIXME: operation not yet implemented: Record::is_subrange_equal")
+      + FILENAME(__LINE__));
   }
 
 }
