@@ -1,15 +1,17 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
+
+#define FILENAME(line) FILENAME_FOR_EXCEPTIONS("src/libawkward/array/UnmaskedArray.cpp", line)
+#define FILENAME_C(line) FILENAME_FOR_EXCEPTIONS_C("src/libawkward/array/UnmaskedArray.cpp", line)
 
 #include <sstream>
 #include <type_traits>
 
-#include "awkward/cpu-kernels/identities.h"
-#include "awkward/cpu-kernels/getitem.h"
-#include "awkward/cpu-kernels/operations.h"
+#include "awkward/kernels.h"
+#include "awkward/kernel-utils.h"
 #include "awkward/type/OptionType.h"
-#include "awkward/type/ArrayType.h"
-#include "awkward/type/UnknownType.h"
+#include "awkward/Reducer.h"
 #include "awkward/Slice.h"
+#include "awkward/io/json.h"
 #include "awkward/array/None.h"
 #include "awkward/array/EmptyArray.h"
 #include "awkward/array/UnionArray.h"
@@ -18,6 +20,7 @@
 #include "awkward/array/ByteMaskedArray.h"
 #include "awkward/array/BitMaskedArray.h"
 #include "awkward/array/VirtualArray.h"
+#include "awkward/array/RegularArray.h"
 
 #include "awkward/array/UnmaskedArray.h"
 
@@ -26,8 +29,9 @@ namespace awkward {
 
   UnmaskedForm::UnmaskedForm(bool has_identities,
                              const util::Parameters& parameters,
+                             const FormKey& form_key,
                              const FormPtr& content)
-      : Form(has_identities, parameters)
+      : Form(has_identities, parameters, form_key)
       , content_(content) { }
 
   const FormPtr
@@ -52,6 +56,7 @@ namespace awkward {
     content_.get()->tojson_part(builder, verbose);
     identities_tojson(builder, verbose);
     parameters_tojson(builder, verbose);
+    form_key_tojson(builder, verbose);
     builder.endrecord();
   }
 
@@ -59,6 +64,15 @@ namespace awkward {
   UnmaskedForm::shallow_copy() const {
     return std::make_shared<UnmaskedForm>(has_identities_,
                                           parameters_,
+                                          form_key_,
+                                          content_);
+  }
+
+  const FormPtr
+  UnmaskedForm::with_form_key(const FormKey& form_key) const {
+    return std::make_shared<UnmaskedForm>(has_identities_,
+                                          parameters_,
+                                          form_key,
                                           content_);
   }
 
@@ -83,6 +97,11 @@ namespace awkward {
     return content_.get()->purelist_depth();
   }
 
+  bool
+  UnmaskedForm::dimension_optiontype() const {
+    return true;
+  }
+
   const std::pair<int64_t, int64_t>
   UnmaskedForm::minmax_depth() const {
     return content_.get()->minmax_depth();
@@ -121,25 +140,77 @@ namespace awkward {
   bool
   UnmaskedForm::equal(const FormPtr& other,
                       bool check_identities,
-                      bool check_parameters) const {
+                      bool check_parameters,
+                      bool check_form_key,
+                      bool compatibility_check) const {
+    if (compatibility_check) {
+      if (VirtualForm* raw = dynamic_cast<VirtualForm*>(other.get())) {
+        if (raw->form().get() != nullptr) {
+          return equal(raw->form(),
+                       check_identities,
+                       check_parameters,
+                       check_form_key,
+                       compatibility_check);
+        }
+      }
+    }
+
     if (check_identities  &&
         has_identities_ != other.get()->has_identities()) {
       return false;
     }
     if (check_parameters  &&
-        !util::parameters_equal(parameters_, other.get()->parameters())) {
+        !util::parameters_equal(parameters_, other.get()->parameters(), false)) {
+      return false;
+    }
+    if (check_form_key  &&
+        !form_key_equals(other.get()->form_key())) {
       return false;
     }
     if (UnmaskedForm* t = dynamic_cast<UnmaskedForm*>(other.get())) {
       return (content_.get()->equal(t->content(),
                                     check_identities,
-                                    check_parameters));
+                                    check_parameters,
+                                    check_form_key,
+                                    compatibility_check));
     }
     else {
       return false;
     }
   }
 
+  const FormPtr
+  UnmaskedForm::getitem_field(const std::string& key) const {
+    UnmaskedForm step1(has_identities_,
+                       util::Parameters(),
+                       FormKey(nullptr),
+                       content_.get()->getitem_field(key));
+    return step1.simplify_optiontype();
+  }
+
+  const FormPtr
+  UnmaskedForm::getitem_fields(const std::vector<std::string>& keys) const {
+    UnmaskedForm step1(has_identities_,
+                       util::Parameters(),
+                       FormKey(nullptr),
+                       content_.get()->getitem_fields(keys));
+    return step1.simplify_optiontype();
+  }
+
+  const FormPtr
+  UnmaskedForm::simplify_optiontype() const {
+    if (dynamic_cast<IndexedForm*>(content_.get())         ||
+        dynamic_cast<IndexedOptionForm*>(content_.get())   ||
+        dynamic_cast<ByteMaskedForm*>(content_.get())      ||
+        dynamic_cast<BitMaskedForm*>(content_.get())       ||
+        dynamic_cast<UnmaskedForm*>(content_.get())) {
+      return content_;
+    }
+    else {
+      return shallow_copy();
+    }
+  }
+
   ////////// UnmaskedArray
 
   UnmaskedArray::UnmaskedArray(const IdentitiesPtr& identities,
@@ -170,8 +241,9 @@ namespace awkward {
   const Index8
   UnmaskedArray::bytemask() const {
     Index8 out(length());
-    struct Error err = awkward_zero_mask8(
-      out.ptr().get(),
+    struct Error err = kernel::zero_mask8(
+      kernel::lib::cpu,   // DERIVE
+      out.data(),
       length());
     util::handle_error(err, classname(), identities_.get());
     return out;
@@ -194,11 +266,28 @@ namespace awkward {
     }
   }
 
+  const std::shared_ptr<ByteMaskedArray>
+  UnmaskedArray::toByteMaskedArray() const {
+    Index8 bytemask(length());
+    struct Error err = kernel::one_mask8(
+      kernel::lib::cpu,   // DERIVE
+      bytemask.data(),
+      length());
+    util::handle_error(err, classname(), identities_.get());
+    return std::make_shared<ByteMaskedArray>(
+      identities_,
+      parameters_,
+      bytemask,
+      content_,
+      true);
+  }
+
   const ContentPtr
   UnmaskedArray::toIndexedOptionArray64() const {
     Index64 index(length());
-    struct Error err = awkward_carry_arange_64(
-      index.ptr().get(),
+    struct Error err = kernel::carry_arange<int64_t>(
+      kernel::lib::cpu,   // DERIVE
+      index.data(),
       length());
     util::handle_error(err, classname(), identities_.get());
     return std::make_shared<IndexedOptionArray64>(identities_,
@@ -222,7 +311,8 @@ namespace awkward {
         util::handle_error(
           failure("content and its identities must have the same length",
                   kSliceNone,
-                  kSliceNone),
+                  kSliceNone,
+                  FILENAME_C(__LINE__)),
           classname(),
           identities_.get());
       }
@@ -235,10 +325,10 @@ namespace awkward {
                                          content_.get()->length());
         Identities32* rawsubidentities =
           reinterpret_cast<Identities32*>(subidentities.get());
-        struct Error err = awkward_identities32_extend(
-          rawsubidentities->ptr().get(),
-          rawidentities->ptr().get(),
-          rawidentities->offset(),
+        struct Error err = kernel::Identities_extend<int32_t>(
+          kernel::lib::cpu,   // DERIVE
+          rawsubidentities->data(),
+          rawidentities->data(),
           rawidentities->length(),
           content_.get()->length());
         util::handle_error(err, classname(), identities_.get());
@@ -253,17 +343,18 @@ namespace awkward {
                                          content_.get()->length());
         Identities64* rawsubidentities =
           reinterpret_cast<Identities64*>(subidentities.get());
-        struct Error err = awkward_identities64_extend(
-          rawsubidentities->ptr().get(),
-          rawidentities->ptr().get(),
-          rawidentities->offset(),
+        struct Error err = kernel::Identities_extend<int64_t>(
+          kernel::lib::cpu,   // DERIVE
+          rawsubidentities->data(),
+          rawidentities->data(),
           rawidentities->length(),
           content_.get()->length());
         util::handle_error(err, classname(), identities_.get());
         content_.get()->setidentities(subidentities);
       }
       else {
-        throw std::runtime_error("unrecognized Identities specialization");
+        throw std::runtime_error(
+          std::string("unrecognized Identities specialization") + FILENAME(__LINE__));
       }
     }
     identities_ = identities;
@@ -279,8 +370,10 @@ namespace awkward {
                                        length());
       Identities32* rawidentities =
         reinterpret_cast<Identities32*>(newidentities.get());
-      struct Error err = awkward_new_identities32(rawidentities->ptr().get(),
-                                                  length());
+      struct Error err = kernel::new_Identities<int32_t>(
+        kernel::lib::cpu,   // DERIVE
+        rawidentities->data(),
+        length());
       util::handle_error(err, classname(), identities_.get());
       setidentities(newidentities);
     }
@@ -292,8 +385,10 @@ namespace awkward {
                                        length());
       Identities64* rawidentities =
         reinterpret_cast<Identities64*>(newidentities.get());
-      struct Error err = awkward_new_identities64(rawidentities->ptr().get(),
-                                                  length());
+      struct Error err = kernel::new_Identities<int64_t>(
+        kernel::lib::cpu,   // DERIVE
+        rawidentities->data(),
+        length());
       util::handle_error(err, classname(), identities_.get());
       setidentities(newidentities);
     }
@@ -309,17 +404,31 @@ namespace awkward {
   UnmaskedArray::form(bool materialize) const {
     return std::make_shared<UnmaskedForm>(identities_.get() != nullptr,
                                           parameters_,
+                                          FormKey(nullptr),
                                           content_.get()->form(materialize));
   }
 
-  bool
-  UnmaskedArray::has_virtual_form() const {
-    return content_.get()->has_virtual_form();
+  kernel::lib
+  UnmaskedArray::kernels() const {
+    if (identities_.get() == nullptr) {
+      return content_.get()->kernels();
+    }
+    else {
+      if (dynamic_cast<EmptyArray*>(content_.get())) {
+        return identities_.get()->ptr_lib();
+      }
+      else if (content_.get()->kernels() == identities_.get()->ptr_lib()) {
+        return identities_.get()->ptr_lib();
+      }
+      else {
+        return kernel::lib::size;
+      }
+    }
   }
 
-  bool
-  UnmaskedArray::has_virtual_length() const {
-    return content_.get()->has_virtual_length();
+  void
+  UnmaskedArray::caches(std::vector<ArrayCachePtr>& out) const {
+    content_.get()->caches(out);
   }
 
   const std::string
@@ -381,7 +490,10 @@ namespace awkward {
     if (identities_.get() != nullptr  &&
         identities_.get()->length() < length()) {
       util::handle_error(
-        failure("len(identities) < len(array)", kSliceNone, kSliceNone),
+        failure("len(identities) < len(array)",
+                kSliceNone,
+                kSliceNone,
+                FILENAME_C(__LINE__)),
         identities_.get()->classname(),
         nullptr);
     }
@@ -400,7 +512,7 @@ namespace awkward {
     }
     if (!(0 <= regular_at  &&  regular_at < length())) {
       util::handle_error(
-        failure("index out of range", kSliceNone, at),
+        failure("index out of range", kSliceNone, at, FILENAME_C(__LINE__)),
         classname(),
         identities_.get());
     }
@@ -416,12 +528,15 @@ namespace awkward {
   UnmaskedArray::getitem_range(int64_t start, int64_t stop) const {
     int64_t regular_start = start;
     int64_t regular_stop = stop;
-    awkward_regularize_rangeslice(&regular_start, &regular_stop,
+    kernel::regularize_rangeslice(&regular_start, &regular_stop,
       true, start != Slice::none(), stop != Slice::none(), length());
     if (identities_.get() != nullptr  &&
         regular_stop > identities_.get()->length()) {
       util::handle_error(
-        failure("index out of range", kSliceNone, stop),
+        failure("index out of range",
+                kSliceNone,
+                stop,
+                FILENAME_C(__LINE__)),
         identities_.get()->classname(),
         nullptr);
     }
@@ -442,18 +557,36 @@ namespace awkward {
 
   const ContentPtr
   UnmaskedArray::getitem_field(const std::string& key) const {
-    return std::make_shared<UnmaskedArray>(
-      identities_,
-      util::Parameters(),
-      content_.get()->getitem_field(key));
+    UnmaskedArray step1(identities_,
+                        util::Parameters(),
+                        content_.get()->getitem_field(key));
+    return step1.simplify_optiontype();
+  }
+
+  const ContentPtr
+  UnmaskedArray::getitem_field(const std::string& key,
+                               const Slice& only_fields) const {
+    UnmaskedArray step1(identities_,
+                        util::Parameters(),
+                        content_.get()->getitem_field(key, only_fields));
+    return step1.simplify_optiontype();
   }
 
   const ContentPtr
   UnmaskedArray::getitem_fields(const std::vector<std::string>& keys) const {
-    return std::make_shared<UnmaskedArray>(
-      identities_,
-      util::Parameters(),
-      content_.get()->getitem_fields(keys));
+    UnmaskedArray step1(identities_,
+                        util::Parameters(),
+                        content_.get()->getitem_fields(keys));
+    return step1.simplify_optiontype();
+  }
+
+  const ContentPtr
+  UnmaskedArray::getitem_fields(const std::vector<std::string>& keys,
+                                const Slice& only_fields) const {
+    UnmaskedArray step1(identities_,
+                        util::Parameters(),
+                        content_.get()->getitem_fields(keys, only_fields));
+    return step1.simplify_optiontype();
   }
 
   const ContentPtr
@@ -492,20 +625,48 @@ namespace awkward {
              dynamic_cast<SliceMissing64*>(head.get())) {
       return Content::getitem_next(*missing, tail, advanced);
     }
+    else if (SliceVarNewAxis* varnewaxis =
+             dynamic_cast<SliceVarNewAxis*>(head.get())) {
+      return UnmaskedArray::getitem_next(*varnewaxis, tail, advanced);
+    }
     else {
-      throw std::runtime_error("unrecognized slice type");
+      throw std::runtime_error(
+        std::string("unrecognized slice type") + FILENAME(__LINE__));
     }
   }
 
   const ContentPtr
-  UnmaskedArray::carry(const Index64& carry) const {
+  UnmaskedArray::carry(const Index64& carry, bool allow_lazy) const {
+    if (carry.iscontiguous()) {
+      if (carry.length() == length()) {
+        return shallow_copy();
+      }
+      else {
+        return getitem_range_nowrap(0, carry.length());
+      }
+    }
     IdentitiesPtr identities(nullptr);
     if (identities_.get() != nullptr) {
       identities = identities_.get()->getitem_carry_64(carry);
     }
     return std::make_shared<UnmaskedArray>(identities,
                                            parameters_,
-                                           content_.get()->carry(carry));
+                                           content_.get()->carry(carry, allow_lazy));
+  }
+
+  int64_t
+  UnmaskedArray::purelist_depth() const {
+    return content_.get()->purelist_depth();
+  }
+
+  const std::pair<int64_t, int64_t>
+  UnmaskedArray::minmax_depth() const {
+    return content_.get()->minmax_depth();
+  }
+
+  const std::pair<bool, int64_t>
+  UnmaskedArray::branch_depth() const {
+    return content_.get()->branch_depth();
   }
 
   int64_t
@@ -535,6 +696,10 @@ namespace awkward {
 
   const std::string
   UnmaskedArray::validityerror(const std::string& path) const {
+    const std::string paramcheck = validityerror_parameters(path);
+    if (paramcheck != std::string("")) {
+      return paramcheck;
+    }
     return content_.get()->validityerror(path + std::string(".content"));
   }
 
@@ -545,8 +710,8 @@ namespace awkward {
 
   const ContentPtr
   UnmaskedArray::num(int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       Index64 out(1);
       out.setitem_at_nowrap(0, length());
       return NumpyArray(out).getitem_at_nowrap(0);
@@ -554,19 +719,20 @@ namespace awkward {
     else {
       return std::make_shared<UnmaskedArray>(Identities::none(),
                                              util::Parameters(),
-                                             content_.get()->num(axis, depth));
+                                             content_.get()->num(posaxis, depth));
     }
   }
 
   const std::pair<Index64, ContentPtr>
   UnmaskedArray::offsets_and_flattened(int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
-      throw std::invalid_argument("axis=0 not allowed for flatten");
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
+      throw std::invalid_argument(
+        std::string("axis=0 not allowed for flatten") + FILENAME(__LINE__));
     }
     else {
       std::pair<Index64, ContentPtr> offsets_flattened =
-        content_.get()->offsets_and_flattened(axis, depth);
+        content_.get()->offsets_and_flattened(posaxis, depth);
       Index64 offsets = offsets_flattened.first;
       ContentPtr flattened = offsets_flattened.second;
       if (offsets.length() == 0) {
@@ -588,7 +754,7 @@ namespace awkward {
       return mergeable(raw->array(), mergebool);
     }
 
-    if (!parameters_equal(other.get()->parameters())) {
+    if (!parameters_equal(other.get()->parameters(), false)) {
       return false;
     }
 
@@ -636,17 +802,39 @@ namespace awkward {
     }
   }
 
-  const ContentPtr
-  UnmaskedArray::reverse_merge(const ContentPtr& other) const {
-    ContentPtr indexedoptionarray = toIndexedOptionArray64();
-    IndexedOptionArray64* raw =
-      dynamic_cast<IndexedOptionArray64*>(indexedoptionarray.get());
-    return raw->reverse_merge(other);
+  bool
+  UnmaskedArray::referentially_equal(const ContentPtr& other) const {
+    if (identities_.get() == nullptr  &&  other.get()->identities().get() != nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() == nullptr) {
+      return false;
+    }
+    if (identities_.get() != nullptr  &&  other.get()->identities().get() != nullptr) {
+      if (!identities_.get()->referentially_equal(other->identities())) {
+        return false;
+      }
+    }
+    if (UnmaskedArray* raw = dynamic_cast<UnmaskedArray*>(other.get())) {
+      return parameters_ == raw->parameters()  &&
+             content_.get()->referentially_equal(raw->content());
+    }
+    else {
+      return false;
+    }
   }
 
   const ContentPtr
-  UnmaskedArray::merge(const ContentPtr& other) const {
-    return toIndexedOptionArray64().get()->merge(other);
+  UnmaskedArray::reverse_merge(const ContentPtr& other) const {
+    return toIndexedOptionArray64().get()->reverse_merge(other);
+  }
+
+  const ContentPtr
+  UnmaskedArray::mergemany(const ContentPtrVec& others) const {
+    if (others.empty()) {
+      return shallow_copy();
+    }
+    return toIndexedOptionArray64().get()->mergemany(others);
   }
 
   const SliceItemPtr
@@ -661,18 +849,18 @@ namespace awkward {
 
   const ContentPtr
   UnmaskedArray::rpad(int64_t target, int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       return rpad_axis0(target, false);
     }
-    else if (toaxis == depth + 1) {
-      return content_.get()->rpad(target, axis, depth);
+    else if (posaxis == depth + 1) {
+      return content_.get()->rpad(target, posaxis, depth);
     }
     else {
       return std::make_shared<UnmaskedArray>(
         Identities::none(),
         parameters_,
-        content_.get()->rpad(target, axis, depth));
+        content_.get()->rpad(target, posaxis, depth));
     }
   }
 
@@ -680,18 +868,18 @@ namespace awkward {
   UnmaskedArray::rpad_and_clip(int64_t target,
                                int64_t axis,
                                int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (toaxis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       return rpad_axis0(target, false);
     }
-    else if (toaxis == depth + 1) {
-      return content_.get()->rpad_and_clip(target, axis, depth);
+    else if (posaxis == depth + 1) {
+      return content_.get()->rpad_and_clip(target, posaxis, depth);
     }
     else {
       return std::make_shared<UnmaskedArray>(
         Identities::none(),
         parameters_,
-        content_.get()->rpad_and_clip(target, axis, depth));
+        content_.get()->rpad_and_clip(target, posaxis, depth));
     }
   }
 
@@ -699,30 +887,36 @@ namespace awkward {
   UnmaskedArray::reduce_next(const Reducer& reducer,
                              int64_t negaxis,
                              const Index64& starts,
+                             const Index64& shifts,
                              const Index64& parents,
                              int64_t outlength,
                              bool mask,
                              bool keepdims) const {
-    return content_.get()->reduce_next(reducer,
-                                       negaxis,
-                                       starts,
-                                       parents,
-                                       outlength,
-                                       mask,
-                                       keepdims);
+    ContentPtr next = content_;
+    if (RegularArray* raw = dynamic_cast<RegularArray*>(next.get())) {
+      next = raw->toListOffsetArray64(true);
+    }
+    return next.get()->reduce_next(reducer,
+                                   negaxis,
+                                   starts,
+                                   shifts,
+                                   parents,
+                                   outlength,
+                                   mask,
+                                   keepdims);
   }
 
   const ContentPtr
   UnmaskedArray::localindex(int64_t axis, int64_t depth) const {
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (axis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       return localindex_axis0();
     }
     else {
       return std::make_shared<UnmaskedArray>(
         identities_,
         util::Parameters(),
-        content_.get()->localindex(axis, depth));
+        content_.get()->localindex(posaxis, depth));
     }
   }
 
@@ -734,10 +928,12 @@ namespace awkward {
                               int64_t axis,
                               int64_t depth) const {
     if (n < 1) {
-      throw std::invalid_argument("in combinations, 'n' must be at least 1");
+      throw std::invalid_argument(
+        std::string("in combinations, 'n' must be at least 1")
+        + FILENAME(__LINE__));
     }
-    int64_t toaxis = axis_wrap_if_negative(axis);
-    if (axis == depth) {
+    int64_t posaxis = axis_wrap_if_negative(axis);
+    if (posaxis == depth) {
       return combinations_axis0(n, replacement, recordlookup, parameters);
     }
     else {
@@ -748,17 +944,88 @@ namespace awkward {
                                      replacement,
                                      recordlookup,
                                      parameters,
-                                     axis,
+                                     posaxis,
                                      depth));
     }
   }
 
+  const ContentPtr
+  UnmaskedArray::sort_next(int64_t negaxis,
+                           const Index64& starts,
+                           const Index64& parents,
+                           int64_t outlength,
+                           bool ascending,
+                           bool stable,
+                           bool keepdims) const {
+    if (length() == 0) {
+      return shallow_copy();
+    }
+    std::shared_ptr<Content> out = content_.get()->sort_next(negaxis,
+                                                             starts,
+                                                             parents,
+                                                             outlength,
+                                                             ascending,
+                                                             stable,
+                                                             keepdims);
+    if (RegularArray* raw = dynamic_cast<RegularArray*>(out.get())) {
+      std::shared_ptr<Content> wrapped = std::make_shared<UnmaskedArray>(
+          Identities::none(),
+          parameters_,
+          raw->content());
+      return std::make_shared<RegularArray>(
+          raw->identities(),
+          raw->parameters(),
+          wrapped,
+          raw->size(),
+          length());
+    }
+    else {
+      return out;
+    }
+  }
+
+  const ContentPtr
+  UnmaskedArray::argsort_next(int64_t negaxis,
+                              const Index64& starts,
+                              const Index64& parents,
+                              int64_t outlength,
+                              bool ascending,
+                              bool stable,
+                              bool keepdims) const {
+    if (length() == 0) {
+      return shallow_copy();
+    }
+    std::shared_ptr<Content> out = content_.get()->argsort_next(negaxis,
+                                                                starts,
+                                                                parents,
+                                                                outlength,
+                                                                ascending,
+                                                                stable,
+                                                                keepdims);
+    if (RegularArray* raw = dynamic_cast<RegularArray*>(out.get())) {
+      std::shared_ptr<Content> wrapped = std::make_shared<UnmaskedArray>(
+          Identities::none(),
+          parameters_,
+          raw->content());
+      return std::make_shared<RegularArray>(
+          raw->identities(),
+          raw->parameters(),
+          wrapped,
+          raw->size(),
+          length());
+    }
+    else {
+      return out;
+    }
+  }
+
   const ContentPtr
   UnmaskedArray::getitem_next(const SliceAt& at,
                               const Slice& tail,
                               const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: UnmaskedArray::getitem_next(at)");
+      std::string("undefined operation: UnmaskedArray::getitem_next(at)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -766,7 +1033,8 @@ namespace awkward {
                               const Slice& tail,
                               const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: UnmaskedArray::getitem_next(range)");
+      std::string("undefined operation: UnmaskedArray::getitem_next(range)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -774,7 +1042,8 @@ namespace awkward {
                               const Slice& tail,
                               const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: UnmaskedArray::getitem_next(array)");
+      std::string("undefined operation: UnmaskedArray::getitem_next(array)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -782,7 +1051,8 @@ namespace awkward {
                               const Slice& tail,
                               const Index64& advanced) const {
     throw std::runtime_error(
-      "undefined operation: UnmaskedArray::getitem_next(jagged)");
+      std::string("undefined operation: UnmaskedArray::getitem_next(jagged)")
+      + FILENAME(__LINE__));
   }
 
   const ContentPtr
@@ -818,6 +1088,52 @@ namespace awkward {
                                                       tail);
   }
 
+  const ContentPtr
+  UnmaskedArray::getitem_next_jagged(const Index64& slicestarts,
+                                     const Index64& slicestops,
+                                     const SliceVarNewAxis& slicecontent,
+                                     const Slice& tail) const {
+    return getitem_next_jagged_generic<SliceVarNewAxis>(slicestarts,
+                                                        slicestops,
+                                                        slicecontent,
+                                                        tail);
+  }
+
+  const ContentPtr
+  UnmaskedArray::getitem_next(const SliceVarNewAxis& varnewaxis,
+                              const Slice& tail,
+                              const Index64& advanced) const {
+    SliceJagged64 jagged = content_.get()->varaxis_to_jagged(varnewaxis);
+    return getitem_next(jagged, tail, advanced);
+  }
+
+  const SliceJagged64
+  UnmaskedArray::varaxis_to_jagged(const SliceVarNewAxis& varnewaxis) const {
+    return content_.get()->varaxis_to_jagged(varnewaxis);
+  }
+
+  const ContentPtr
+  UnmaskedArray::copy_to(kernel::lib ptr_lib) const {
+    ContentPtr content = content_.get()->copy_to(ptr_lib);
+    IdentitiesPtr identities(nullptr);
+    if (identities_.get() != nullptr) {
+      identities = identities_.get()->copy_to(ptr_lib);
+    }
+    return std::make_shared<UnmaskedArray>(identities,
+                                           parameters_,
+                                           content);
+  }
+
+  const ContentPtr
+  UnmaskedArray::numbers_to_type(const std::string& name) const {
+    ContentPtr content = content_.get()->numbers_to_type(name);
+    IdentitiesPtr identities = identities_;
+    if (identities_.get() != nullptr) {
+      identities = identities_.get()->deep_copy();
+    }
+    return std::make_shared<UnmaskedArray>(identities, parameters_, content);
+  }
+
   template <typename S>
   const ContentPtr
   UnmaskedArray::getitem_next_jagged_generic(const Index64& slicestarts,
@@ -833,4 +1149,26 @@ namespace awkward {
     return out2.simplify_optiontype();
   }
 
+  bool
+  UnmaskedArray::is_unique() const {
+    throw std::runtime_error(
+      std::string("FIXME: operation not yet implemented: UnmaskedArray::is_unique")
+      + FILENAME(__LINE__));
+
+  }
+
+  const ContentPtr
+  UnmaskedArray::unique() const {
+    throw std::runtime_error(
+      std::string("FIXME: operation not yet implemented: UnmaskedArray::unique")
+      + FILENAME(__LINE__));
+  }
+
+  bool
+  UnmaskedArray::is_subrange_equal(const Index64& start, const Index64& stop) const {
+    throw std::runtime_error(
+      std::string("FIXME: operation not yet implemented: UnmaskedArray::is_subrange_equal")
+      + FILENAME(__LINE__));
+  }
+
 }
