@@ -1,4 +1,6 @@
-// BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE
+// BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
+
+#define FILENAME(line) FILENAME_FOR_EXCEPTIONS("src/libawkward/type/ArrayType.cpp", line)
 
 #include <string>
 
@@ -33,7 +35,7 @@ namespace awkward {
   bool
   ArrayType::equal(const TypePtr& other, bool check_parameters) const {
     if (ArrayType* t = dynamic_cast<ArrayType*>(other.get())) {
-      if (check_parameters  &&  !parameters_equal(other.get()->parameters())) {
+      if (check_parameters  &&  !parameters_equal(other.get()->parameters(), false)) {
         return false;
       }
       return (length_ == t->length_  &&
@@ -74,7 +76,7 @@ namespace awkward {
     if (length_ != 0) {
       throw std::invalid_argument(
         std::string("ArrayType with length ") + std::to_string(length_)
-        + std::string(" does not describe an empty array"));
+        + std::string(" does not describe an empty array") + FILENAME(__LINE__));
     }
     return type_.get()->empty();
   }
